"""
Miscellaneous utilities used for development.

Copyright by Gabriel A. Hackebeil (gabe.hackebeil@gmail.com).
"""
import math
import array
import collections
import hashlib

import pybnb

import six
from six.moves import xrange as range

import pyomo.kernel as pmo
if getattr(pmo,'version_info',(0,)*3) < (5,4,3):   #pragma:nocover
    raise ImportError(
        "Pyomo 5.4.3 or later is not available")

def hash_joblist(jobs):
    """Create a hash of a Python list by casting each entry
    to a string."""
    x = hashlib.sha1()
    for entry in jobs:
        x.update(str(entry).encode())
    return x.hexdigest()

def add_tmp_component(model, name, obj):
    """Add a temporary component to a model, adjusting the
    name as needed to make sure it is unique."""
    while hasattr(model, name):
        name = "."+name+"."
    setattr(model, name, obj)
    return name

def mpi_partition(comm, items, root=0):
    """A generator that partitions the list of items across
    processes in the communicator. If the communicator size
    is greater than 1, the root process iterates over no
    items, but rather serves them dynamically after
    receiving requests from workers. This function assumes
    each process has an identical copy of the items
    list. Therefore, items in the list are not transferred
    (only indices)."""
    assert root >= 0
    N = len(items)
    if N > 0:
        if (comm is None) or \
           (comm.size == 1):
            assert root == 0
            for x in items:
                yield x
        else:
            import mpi4py.MPI
            # it would be pretty easy to refactor this
            # code to avoid this limitation
            assert N <= mpi4py.MPI.COMM_WORLD.Get_attr(
                mpi4py.MPI.TAG_UB)
            _null = [array.array('b',[]),mpi4py.MPI.CHAR]
            last_tag = {}
            if comm.rank == root:
                i = 0
                requests = []
                for dest in range(comm.size):
                    if dest == root:
                        continue
                    last_tag[dest] = i
                    requests.append(comm.Isend(_null, dest, tag=i))
                    i += 1
                status = mpi4py.MPI.Status()
                while i < N:
                    comm.Recv(_null, status=status)
                    last_tag[status.Get_source()] = i
                    requests.append(comm.Isend(_null,
                                               status.Get_source(),
                                               tag=i))
                    i += 1
                for dest in last_tag:
                    if last_tag[dest] < N:
                        requests.append(comm.Isend(_null,
                                                   dest,
                                                   tag=N))
                    requests.append(comm.Irecv(_null, dest))
                mpi4py.MPI.Request.Waitall(requests)
            else:
                status = mpi4py.MPI.Status()
                comm.Recv(_null, source=root, status=status)
                if status.Get_tag() >= N:
                    comm.Send(_null,root)
                else:
                    while status.Get_tag() < N:
                        yield items[status.Get_tag()]
                        comm.Sendrecv(_null,
                                      root,
                                      recvbuf=_null,
                                      source=root,
                                      status=status)

def correct_integer_lb(lb,
                       integer_tolerance):
    """Converts a lower bound for an integer optimization
    variable to an integer equal to `ceil(ub)`, taking care
    not to move a non-integer bound away from an integer
    point already within a given tolerance."""
    assert 0 <= integer_tolerance < 0.5
    if lb-math.floor(lb) > integer_tolerance:
        return int(math.ceil(lb))
    else:
        return int(math.floor(lb))

def correct_integer_ub(ub,
                       integer_tolerance):
    """Converts an upper bound for an integer optimization
    variable to an integer equal to `floor(ub)`, taking care
    not to move a non-integer bound away from an integer
    point already within a given tolerance."""
    assert 0 <= integer_tolerance < 0.5
    if math.ceil(ub)-ub > integer_tolerance:
        return int(math.floor(ub))
    else:
        return int(math.ceil(ub))

def create_optimality_bound(problem,
                            pyomo_objective,
                            best_objective_value):
    """Returns a constraint that bounds an objective
    function with a known best value. That is, the
    constraint will require the objective function to be
    better than the given value."""
    optbound = pmo.constraint(body=pyomo_objective)
    if problem.sense() == pybnb.minimize:
        assert pyomo_objective.sense == pmo.minimize
        optbound.ub = best_objective_value
    else:
        assert problem.sense() == pybnb.maximize
        assert pyomo_objective.sense == pmo.maximize
        optbound.lb = best_objective_value
    return optbound

def generate_cids(model,
                  prefix=(),
                  **kwds):
    """Generate forward and reverse mappings between model
    components and deterministic, unique identifiers that
    are safe to serialize or use as dictionary keys."""
    object_to_cid = pmo.ComponentMap()
    cid_to_object = collections.OrderedDict()
    if hasattr(pmo, 'preorder_traversal'):        #pragma:nocover
        fn = lambda *args, **kwds: pmo.preorder_traversal(model,
                                                          *args,
                                                          **kwds)
    else:                                         #pragma:nocover
        fn = model.preorder_traversal
    try:
        fn(return_key=True)
    except TypeError:
        traversal = fn(**kwds)
        obj_ = six.next(traversal)
        assert obj_ is model
        object_to_cid[model] = prefix
        cid_to_object[prefix] = model
        for obj in traversal:
            parent = obj.parent
            key = obj.storage_key
            cid_ = object_to_cid[obj] = object_to_cid[parent]+(key,)
            cid_to_object[cid_] = obj
    else:                                         #pragma:nocover
        traversal = fn(return_key=True, **kwds)
        obj_ = six.next(traversal)[1]
        assert obj_ is model
        object_to_cid[model] = prefix
        cid_to_object[prefix] = model
        for key, obj in traversal:
            parent = obj.parent
            cid_ = object_to_cid[obj] = object_to_cid[parent]+(key,)
            cid_to_object[cid_] = obj
    return object_to_cid, cid_to_object



import os
import sys

from runtests.mpi import MPITest

from pybnb import (QueueStrategy,
                   Node,
                   inf,
                   SolverResults,
                   Solver)
from pybnb.misc import get_simple_logger

from .common import mpi_available

thisdir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, thisdir)
try:
    from problems import (infeasible_max,
                          infeasible_min,
                          root_infeasible_max,
                          root_infeasible_min,
                          unbounded_max,
                          unbounded_min,
                          zero_objective_max,
                          zero_objective_min,
                          delayed_unbounded_max,
                          delayed_unbounded_min)

finally:
    sys.path.remove(thisdir)

_ignore_value_ = object()

_queue_strategies = sorted(QueueStrategy)
_queue_strategies.append((QueueStrategy.bound,
                          QueueStrategy.objective,
                          'breadth'))

import six

def _execute_single_test(problem,
                         baseline,
                         solver=None,
                         comm=None,
                         **kwds):
    if solver is None:
        solver = Solver(comm=comm)
    else:
        assert comm is None
    if solver.comm is not None:
        if solver.comm.rank == 0:
            pass
        elif solver.comm.rank == 1:
            pass
        elif solver.comm.rank == 3:
            pass
    orig = Node()
    problem.save_state(orig)
    results = solver.solve(problem, **kwds)
    assert isinstance(results.solution_status,
                      six.string_types)
    assert isinstance(results.termination_condition,
                      six.string_types)
    current = Node()
    problem.save_state(current)
    assert current.state == orig.state
    assert len(vars(results)) > 0
    assert len(vars(results)) == len(vars(baseline))
    for name in sorted(sorted(list(vars(results).keys()))):
        if getattr(baseline, name) is _ignore_value_:
            continue
        if (name == 'nodes') and \
           (solver.comm is not None) and \
           (solver.comm.size > 2):
            pass
        else:
            assert getattr(results, name) == getattr(baseline, name), \
                ("value for '"+str(name)+"' ("+
                 str(getattr(results, name))+") does "
                 "not match baseline ("+
                 str(getattr(baseline, name))+")")
    if solver.is_dispatcher:
        q = solver.save_dispatcher_queue()
        assert len(q.nodes) == solver._disp.queue.size()
        assert q.sense == solver._disp.converger.sense
        assert q.worst_terminal_bound == solver._disp.worst_terminal_bound
        assert q.bound() == results.bound

def _execute_tests(comm, problem, baseline, **kwds):
    assert 'log_interval_second' not in kwds
    assert 'log' not in kwds
    _execute_single_test(problem, baseline, comm=comm, **kwds)
    kwds['log_interval_seconds'] = 0.0
    _execute_single_test(problem, baseline, comm=comm, **kwds)
    kwds['log_interval_seconds'] = 100.0
    _execute_single_test(problem, baseline, comm=comm, **kwds)
    kwds['log_interval_seconds'] = 0.0
    kwds['log'] = None
    _execute_single_test(problem, baseline, comm=comm, **kwds)
    kwds['log_interval_seconds'] = 100.0
    _execute_single_test(problem, baseline, comm=comm, **kwds)
    solver = Solver(comm=comm)
    _execute_single_test(problem, baseline, solver=solver, **kwds)
    _execute_single_test(problem, baseline, solver=solver, **kwds)
    kwds['log'] = get_simple_logger(level="WARNING")
    _execute_single_test(problem, baseline, solver=solver, **kwds)

def _test_infeasible_max(comm):
    solver = None
    if comm is not None:
        solver = Solver(comm=comm)
    baseline = SolverResults()
    baseline.solution_status = "infeasible"
    baseline.termination_condition = "optimality"
    baseline.objective = -inf
    baseline.bound = -inf
    baseline.nodes = 255
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = infeasible_max.InfeasibleMax()
    if comm is None:
        _execute_tests(comm,
                       problem,
                       baseline)
    else:
        _execute_single_test(problem,
                             baseline,
                             solver=solver)


    baseline = SolverResults()
    baseline.solution_status = "unknown"
    baseline.termination_condition = "no_nodes"
    baseline.objective = -inf
    baseline.bound = -16
    baseline.nodes = 31
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = infeasible_max.InfeasibleMax(branching_abstol=0.1)
    if comm is None:
        _execute_tests(comm,
                       problem,
                       baseline)
    else:
        _execute_single_test(problem,
                             baseline,
                             solver=solver)

    baseline = SolverResults()
    baseline.solution_status = "unknown"
    baseline.termination_condition = "objective_limit"
    baseline.objective = -inf
    baseline.bound = -16
    baseline.nodes = 31
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = infeasible_max.InfeasibleMax(branching_abstol=0.1)
    if comm is None:
        _execute_tests(comm,
                       problem,
                       baseline,
                       bound_stop=-15)
    else:
        _execute_single_test(problem,
                             baseline,
                             solver=solver,
                             bound_stop=-15)

    baseline = SolverResults()
    baseline.solution_status = "feasible"
    baseline.termination_condition = "objective_limit"
    baseline.objective = -20
    baseline.bound = -16
    baseline.absolute_gap = 4
    baseline.relative_gap = 0.2
    baseline.nodes = 31
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = infeasible_max.InfeasibleMax(branching_abstol=0.1,
                                           fixed_objective=-20)
    if comm is None:
        _execute_tests(comm,
                       problem,
                       baseline,
                       bound_stop=-15)
    else:
        _execute_single_test(problem,
                             baseline,
                             solver=solver,
                             bound_stop=-15)

    baseline = SolverResults()
    baseline.solution_status = "unknown"
    baseline.termination_condition = "node_limit"
    baseline.objective = -inf
    baseline.bound = -16
    baseline.nodes = 31
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = infeasible_max.InfeasibleMax()
    if comm is None:
        _execute_tests(comm,
                       problem,
                       baseline,
                       node_limit=31)
    elif comm.size <= 2:
        # skip for larger comm sizes
        # as the node_limit can lead to
        # a number of different outcomes
        _execute_single_test(problem,
                             baseline,
                             solver=solver,
                             node_limit=31)

    baseline = SolverResults()
    baseline.solution_status = "feasible"
    baseline.termination_condition = "node_limit"
    baseline.objective = -17
    baseline.bound = -16
    baseline.absolute_gap = 1.0
    baseline.relative_gap = 1.0/17
    baseline.nodes = 31
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = infeasible_max.InfeasibleMax(fixed_objective=-17)
    if comm is None:
        _execute_tests(comm,
                       problem,
                       baseline,
                       node_limit=31)
    elif comm.size <= 2:
        # skip for larger comm sizes
        # as the node_limit can lead to
        # a number of different outcomes
        _execute_single_test(problem,
                             baseline,
                             solver=solver,
                             node_limit=31)

    baseline = SolverResults()
    baseline.solution_status = "unknown"
    baseline.termination_condition = "time_limit"
    baseline.objective = -inf
    baseline.bound = inf
    baseline.nodes = 0
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = infeasible_max.InfeasibleMax()
    if comm is None:
        _execute_tests(comm,
                       problem,
                       baseline,
                       time_limit=0)
    else:
        _execute_single_test(problem,
                             baseline,
                             solver=solver,
                             time_limit=0)

    if solver is None:
        solver = Solver(comm=comm)
    baseline = SolverResults()
    baseline.solution_status = "infeasible"
    baseline.termination_condition = "optimality"
    baseline.objective = -inf
    baseline.bound = -inf
    baseline.nodes = 255
    baseline.best_node = None
    baseline.wall_time = _ignore_value_
    problem = infeasible_max.InfeasibleMax()
    for queue_strategy in _queue_strategies:
        if queue_strategy == "custom":
            continue
        _execute_single_test(
            problem,
            baseline,
            solver=solver,
            queue_strategy=queue_strategy)

    baseline = SolverResults()
    baseline.solution_status = "invalid"
    baseline.termination_condition = "optimality"
    baseline.objective = -100000000
    baseline.bound = -inf
    baseline.nodes = _ignore_value_
    baseline.best_node = None
    baseline.wall_time = _ignore_value_
    problem = infeasible_max.InfeasibleMax()
    for queue_strategy in _queue_strategies:
        if queue_strategy == "custom":
            continue
        _execute_single_test(
            problem,
            baseline,
            solver=solver,
            queue_strategy=queue_strategy,
            best_objective=-100000000)

def _test_infeasible_min(comm):
    solver = None
    if comm is not None:
        solver = Solver(comm=comm)
    baseline = SolverResults()
    baseline.solution_status = "infeasible"
    baseline.termination_condition = "optimality"
    baseline.objective = inf
    baseline.bound = inf
    baseline.nodes = 255
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = infeasible_min.InfeasibleMin()
    if comm is None:
        _execute_tests(comm,
                       problem,
                       baseline)
    else:
        _execute_single_test(problem,
                             baseline,
                             solver=solver)

    baseline = SolverResults()
    baseline.solution_status = "unknown"
    baseline.termination_condition = "no_nodes"
    baseline.objective = inf
    baseline.bound = 16
    baseline.nodes = 31
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = infeasible_min.InfeasibleMin(branching_abstol=0.1)
    if comm is None:
        _execute_tests(comm,
                       problem,
                       baseline)
    else:
        _execute_single_test(problem,
                             baseline,
                             solver=solver)

    baseline = SolverResults()
    baseline.solution_status = "unknown"
    baseline.termination_condition = "objective_limit"
    baseline.objective = inf
    baseline.bound = 16
    baseline.nodes = 31
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = infeasible_min.InfeasibleMin(branching_abstol=0.1)
    if comm is None:
        _execute_tests(comm,
                       problem,
                       baseline,
                       bound_stop=15)
    else:
        _execute_single_test(problem,
                             baseline,
                             solver=solver,
                             bound_stop=15)

    baseline = SolverResults()
    baseline.solution_status = "feasible"
    baseline.termination_condition = "objective_limit"
    baseline.objective = 20
    baseline.bound = 16
    baseline.absolute_gap = 4
    baseline.relative_gap = 0.2
    baseline.nodes = 31
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = infeasible_min.InfeasibleMin(branching_abstol=0.1,
                                           fixed_objective=20)
    if comm is None:
        _execute_tests(comm,
                       problem,
                       baseline,
                       bound_stop=15)
    else:
        _execute_single_test(problem,
                             baseline,
                             solver=solver,
                             bound_stop=15)

    baseline = SolverResults()
    baseline.solution_status = "unknown"
    baseline.termination_condition = "node_limit"
    baseline.objective = inf
    baseline.bound = 16
    baseline.nodes = 31
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = infeasible_min.InfeasibleMin()
    if comm is None:
        _execute_tests(comm,
                       problem,
                       baseline,
                       node_limit=31)
    elif comm.size <= 2:
        # skip for larger comm sizes
        # as the node_limit can lead to
        # a number of different outcomes
        _execute_single_test(problem,
                             baseline,
                             solver=solver,
                             node_limit=31)

    baseline = SolverResults()
    baseline.solution_status = "feasible"
    baseline.termination_condition = "node_limit"
    baseline.objective = 17
    baseline.bound = 16
    baseline.absolute_gap = 1
    baseline.relative_gap = 1.0/17
    baseline.nodes = 31
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = infeasible_min.InfeasibleMin(fixed_objective=17)
    if comm is None:
        _execute_tests(comm,
                       problem,
                       baseline,
                       node_limit=31)
    elif comm.size <= 2:
        # skip for larger comm sizes
        # as the node_limit can lead to
        # a number of different outcomes
        _execute_single_test(problem,
                             baseline,
                             solver=solver,
                             node_limit=31)

    baseline = SolverResults()
    baseline.solution_status = "unknown"
    baseline.termination_condition = "time_limit"
    baseline.objective = inf
    baseline.bound = -inf
    baseline.nodes = 0
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = infeasible_min.InfeasibleMin()
    if comm is None:
        _execute_tests(comm,
                       problem,
                       baseline,
                       time_limit=0)
    else:
        _execute_single_test(problem,
                             baseline,
                             solver=solver,
                             time_limit=0)

    if solver is None:
        solver = Solver(comm=comm)
    baseline = SolverResults()
    baseline.solution_status = "infeasible"
    baseline.termination_condition = "optimality"
    baseline.objective = inf
    baseline.bound = inf
    baseline.nodes = 255
    baseline.best_node = None
    baseline.wall_time = _ignore_value_
    problem = infeasible_min.InfeasibleMin()
    for queue_strategy in _queue_strategies:
        if queue_strategy == "custom":
            continue
        _execute_single_test(
            problem,
            baseline,
            solver=solver,
            queue_strategy=queue_strategy)

    baseline = SolverResults()
    baseline.solution_status = "invalid"
    baseline.termination_condition = "optimality"
    baseline.objective = 100000000
    baseline.bound = inf
    baseline.nodes = _ignore_value_
    baseline.best_node = None
    baseline.wall_time = _ignore_value_
    problem = infeasible_min.InfeasibleMin()
    for queue_strategy in _queue_strategies:
        if queue_strategy == "custom":
            continue
        _execute_single_test(
            problem,
            baseline,
            solver=solver,
            queue_strategy=queue_strategy,
            best_objective=100000000)

def _test_root_infeasible_max(comm):
    solver = Solver(comm=comm)
    baseline = SolverResults()
    baseline.solution_status = "infeasible"
    baseline.termination_condition = "optimality"
    baseline.objective = -inf
    baseline.bound = -inf
    baseline.nodes = 1
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = root_infeasible_max.RootInfeasibleMax()
    _execute_single_test(problem,
                         baseline,
                         solver=solver)
    for queue_strategy in _queue_strategies:
        if queue_strategy == "custom":
            continue
        _execute_single_test(
            problem,
            baseline,
            solver=solver,
            queue_strategy=queue_strategy)

def _test_root_infeasible_min(comm):
    solver = Solver(comm=comm)
    baseline = SolverResults()
    baseline.solution_status = "infeasible"
    baseline.termination_condition = "optimality"
    baseline.objective = inf
    baseline.bound = inf
    baseline.nodes = 1
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = root_infeasible_min.RootInfeasibleMin()
    _execute_single_test(problem,
                         baseline,
                         solver=solver)
    for queue_strategy in _queue_strategies:
        if queue_strategy == "custom":
            continue
        _execute_single_test(
            problem,
            baseline,
            solver=solver,
            queue_strategy=queue_strategy)

def _test_unbounded_max(comm):
    solver = Solver(comm=comm)
    baseline = SolverResults()
    baseline.solution_status = "unbounded"
    baseline.termination_condition = "no_nodes"
    baseline.objective = inf
    baseline.bound = inf
    baseline.nodes = 1
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = unbounded_max.UnboundedMax()
    _execute_single_test(problem,
                         baseline,
                         solver=solver)
    for queue_strategy in _queue_strategies:
        if queue_strategy == "custom":
            continue
        _execute_single_test(
            problem,
            baseline,
            solver=solver,
            queue_strategy=queue_strategy)

def _test_unbounded_min(comm):
    solver = Solver(comm=comm)
    baseline = SolverResults()
    baseline.solution_status = "unbounded"
    baseline.termination_condition = "no_nodes"
    baseline.objective = -inf
    baseline.bound = -inf
    baseline.nodes = 1
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = unbounded_min.UnboundedMin()
    _execute_single_test(problem,
                         baseline,
                         solver=solver)
    for queue_strategy in _queue_strategies:
        if queue_strategy == "custom":
            continue
        _execute_single_test(
            problem,
            baseline,
            solver=solver,
            queue_strategy=queue_strategy)

def _test_zero_objective_max(comm):
    solver = Solver(comm=comm)
    baseline = SolverResults()
    baseline.solution_status = "optimal"
    baseline.termination_condition = "optimality"
    baseline.objective = 0.0
    baseline.bound = 0.0078125
    baseline.absolute_gap = 0.0078125
    baseline.relative_gap = 0.0078125
    baseline.nodes = 255
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = zero_objective_max.ZeroObjectiveMax()
    _execute_single_test(problem,
                         baseline,
                         solver=solver,
                         relative_gap=0.01,
                         absolute_gap=0.01)
    _execute_single_test(problem,
                         baseline,
                         solver=solver,
                         relative_gap=0.0,
                         absolute_gap=0.01)
    _execute_single_test(problem,
                         baseline,
                         solver=solver,
                         relative_gap=0.01,
                         absolute_gap=0.0)

    baseline = SolverResults()
    baseline.solution_status = "optimal"
    baseline.termination_condition = "optimality"
    baseline.objective = 0.0
    baseline.bound = 0.0009765625
    baseline.absolute_gap = 0.0009765625
    baseline.relative_gap = 0.0009765625
    baseline.nodes = 2047
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = zero_objective_max.ZeroObjectiveMax()
    _execute_single_test(problem,
                         baseline,
                         solver=solver,
                         relative_gap=0.001,
                         absolute_gap=0.001)
    _execute_single_test(problem,
                         baseline,
                         solver=solver,
                         relative_gap=0.0,
                         absolute_gap=0.001)
    _execute_single_test(problem,
                         baseline,
                         solver=solver,
                         relative_gap=0.001,
                         absolute_gap=0.0)

    baseline = SolverResults()
    baseline.solution_status = "feasible"
    baseline.termination_condition = "no_nodes"
    baseline.objective = 0.0
    baseline.bound = 0.0009765625
    baseline.absolute_gap = 0.0009765625
    baseline.relative_gap = 0.0009765625
    baseline.nodes = 2047
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = zero_objective_max.ZeroObjectiveMax()
    _execute_single_test(problem,
                         baseline,
                         solver=solver,
                         relative_gap=0.0001,
                         absolute_gap=0.0001)
    _execute_single_test(problem,
                         baseline,
                         solver=solver,
                         relative_gap=0.0,
                         absolute_gap=0.0001)
    _execute_single_test(problem,
                         baseline,
                         solver=solver,
                         relative_gap=0.0001,
                         absolute_gap=0.0)

    baseline = SolverResults()
    baseline.solution_status = "optimal"
    baseline.termination_condition = "optimality"
    baseline.objective = 0.0
    baseline.bound = 0.0078125
    baseline.absolute_gap = 0.0078125
    baseline.relative_gap = 0.0078125
    baseline.nodes = None
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = zero_objective_max.ZeroObjectiveMax()
    for queue_strategy in _queue_strategies:
        if queue_strategy == "custom":
            continue
        if queue_strategy in ("depth","lifo"):
            baseline.nodes = 2033
        elif queue_strategy == "random":
            baseline.nodes = _ignore_value_
        else:
            baseline.nodes = 255
        _execute_single_test(
            problem,
            baseline,
            solver=solver,
            relative_gap=0.01,
            absolute_gap=0.01,
            queue_strategy=queue_strategy)

def _test_zero_objective_min(comm):
    solver = Solver(comm=comm)
    baseline = SolverResults()
    baseline.solution_status = "optimal"
    baseline.termination_condition = "optimality"
    baseline.objective = 0.0
    baseline.bound = -0.0078125
    baseline.absolute_gap = 0.0078125
    baseline.relative_gap = 0.0078125
    baseline.nodes = 255
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = zero_objective_min.ZeroObjectiveMin()
    _execute_single_test(problem,
                         baseline,
                         solver=solver,
                         relative_gap=0.01,
                         absolute_gap=0.01)
    _execute_single_test(problem,
                         baseline,
                         solver=solver,
                         relative_gap=0.0,
                         absolute_gap=0.01)
    _execute_single_test(problem,
                         baseline,
                         solver=solver,
                         relative_gap=0.01,
                         absolute_gap=0.0)

    baseline = SolverResults()
    baseline.solution_status = "optimal"
    baseline.termination_condition = "optimality"
    baseline.objective = 0.0
    baseline.bound = -0.0009765625
    baseline.absolute_gap = 0.0009765625
    baseline.relative_gap = 0.0009765625
    baseline.nodes = 2047
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = zero_objective_min.ZeroObjectiveMin()
    _execute_single_test(problem,
                         baseline,
                         solver=solver,
                         relative_gap=0.001,
                         absolute_gap=0.001)
    _execute_single_test(problem,
                         baseline,
                         solver=solver,
                         relative_gap=0.0,
                         absolute_gap=0.001)
    _execute_single_test(problem,
                         baseline,
                         solver=solver,
                         relative_gap=0.001,
                         absolute_gap=0.0)

    baseline = SolverResults()
    baseline.solution_status = "feasible"
    baseline.termination_condition = "no_nodes"
    baseline.objective = 0.0
    baseline.bound = -0.0009765625
    baseline.absolute_gap = 0.0009765625
    baseline.relative_gap = 0.0009765625
    baseline.nodes = 2047
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = zero_objective_min.ZeroObjectiveMin()
    _execute_single_test(problem,
                         baseline,
                         solver=solver,
                         relative_gap=0.0001,
                         absolute_gap=0.0001)
    _execute_single_test(problem,
                         baseline,
                         solver=solver,
                         relative_gap=0.0,
                         absolute_gap=0.0001)
    _execute_single_test(problem,
                         baseline,
                         solver=solver,
                         relative_gap=0.0001,
                         absolute_gap=0.0)

    baseline = SolverResults()
    baseline.solution_status = "optimal"
    baseline.termination_condition = "optimality"
    baseline.objective = 0.0
    baseline.bound = -0.0078125
    baseline.absolute_gap = 0.0078125
    baseline.relative_gap = 0.0078125
    baseline.nodes = None
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = zero_objective_min.ZeroObjectiveMin()
    for queue_strategy in _queue_strategies:
        if queue_strategy == "custom":
            continue
        if queue_strategy in ("depth","lifo"):
            baseline.nodes = 2033
        elif queue_strategy == "random":
            baseline.nodes = _ignore_value_
        else:
            baseline.nodes = 255
        _execute_single_test(
            problem,
            baseline,
            solver=solver,
            relative_gap=0.01,
            absolute_gap=0.01,
            queue_strategy=queue_strategy)

def _test_delayed_unbounded_max(comm):
    solver = Solver(comm=comm)
    baseline = SolverResults()
    baseline.solution_status = "unbounded"
    baseline.termination_condition = "no_nodes"
    baseline.objective = inf
    baseline.bound = inf
    baseline.nodes = _ignore_value_
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = delayed_unbounded_max.DelayedUnboundedMax()
    _execute_single_test(problem,
                         baseline,
                         solver=solver)
    for queue_strategy in _queue_strategies:
        if queue_strategy == "custom":
            continue
        _execute_single_test(
            problem,
            baseline,
            solver=solver,
            queue_strategy=queue_strategy)

def _test_delayed_unbounded_min(comm):
    solver = Solver(comm=comm)
    baseline = SolverResults()
    baseline.solution_status = "unbounded"
    baseline.termination_condition = "no_nodes"
    baseline.objective = -inf
    baseline.bound = -inf
    baseline.nodes = _ignore_value_
    baseline.best_node = _ignore_value_
    baseline.wall_time = _ignore_value_
    problem = delayed_unbounded_min.DelayedUnboundedMin()
    _execute_single_test(problem,
                         baseline,
                         solver=solver)
    for queue_strategy in _queue_strategies:
        if queue_strategy == "custom":
            continue
        _execute_single_test(
            problem,
            baseline,
            solver=solver,
            queue_strategy=queue_strategy)

def test_infeasible_max_nocomm():
    _test_infeasible_max(None)

def test_infeasible_min_nocomm():
    _test_infeasible_min(None)

def test_root_infeasible_max_nocomm():
    _test_root_infeasible_max(None)

def test_root_infeasible_min_nocomm():
    _test_root_infeasible_min(None)

def test_unbounded_max_nocomm():
    _test_unbounded_max(None)

def test_unbounded_min_nocomm():
    _test_unbounded_min(None)

def test_zero_objective_max_nocomm():
    _test_zero_objective_max(None)

def test_zero_objective_min_nocomm():
    _test_zero_objective_min(None)

def test_delayed_unbounded_max_nocomm():
    _test_delayed_unbounded_max(None)

def test_delayed_unbounded_min_nocomm():
    _test_delayed_unbounded_min(None)

if mpi_available:

    @MPITest(commsize=[1, 2, 4])
    def test_infeasible_max(comm):
        _test_infeasible_max(comm)

    @MPITest(commsize=[1, 2, 4])
    def test_infeasible_min(comm):
        _test_infeasible_min(comm)

    @MPITest(commsize=[1, 2, 4])
    def test_root_infeasible_max(comm):
        _test_root_infeasible_max(comm)

    @MPITest(commsize=[1, 2, 4])
    def test_root_infeasible_min(comm):
        _test_root_infeasible_min(comm)

    @MPITest(commsize=[1, 2, 4])
    def test_unbounded_max(comm):
        _test_unbounded_max(comm)

    @MPITest(commsize=[1, 2, 4])
    def test_unbounded_min(comm):
        _test_unbounded_min(comm)

    @MPITest(commsize=[1, 2, 4])
    def test_zero_objective_max(comm):
        _test_zero_objective_max(comm)

    @MPITest(commsize=[1, 2, 4])
    def test_zero_objective_min(comm):
        _test_zero_objective_min(comm)

    @MPITest(commsize=[1, 2, 4])
    def test_delayed_unbounded_max(comm):
        _test_delayed_unbounded_max(comm)

    @MPITest(commsize=[1, 2, 4])
    def test_delayed_unbounded_min(comm):
        _test_delayed_unbounded_min(comm)



import pybnb

class InfeasibleMax(pybnb.Problem):

    def __init__(self,
                 branching_abstol=0.001,
                 fixed_objective=None):
        if fixed_objective is None:
            fixed_objective = self.infeasible_objective()
        assert branching_abstol > 0
        self._branching_abstol = branching_abstol
        self._fixed_objective = fixed_objective
        self._xL = 0
        self._xU = 1

    #
    # Implement Problem abstract methods
    #

    def sense(self):
        return pybnb.maximize

    def objective(self):
        return self._fixed_objective

    def bound(self):
        delta = self._xU - self._xL
        if delta < 0.01:
            return self.infeasible_objective()
        else:
            return -1.0/delta

    def save_state(self, node):
        node.state = (self._xL, self._xU)

    def load_state(self, node):
        (self._xL, self._xU) = node.state

    def branch(self):
        xL, xU = self._xL, self._xU
        if (xU - xL) <= self._branching_abstol:
            return
        mid = 0.5 * (xL + xU)
        child = pybnb.Node()
        child.state = (xL, mid)
        yield child
        child = pybnb.Node()
        child.state = (mid, xU)
        yield child



"""
Various utility function for MPI.

Copyright by Gabriel A. Hackebeil (gabe.hackebeil@gmail.com).
"""

import array

# avoids generating a deprecation warning in python 3.7
def _array_to_string(out):
    """converts an array of bytes to a string"""
    if hasattr(out, 'tobytes'):
        # array.tobytes was added in python 3.2
        return out.tobytes().decode("utf8")
    else:
        return out.tostring().decode("utf8")

class Message(object):
    """A helper class for probing for and receiving
    messages. A single instance of this class is meant to be
    reused.

    Parameters
    ----------
    comm : :class:`mpi4py.MPI.Comm`
        The MPI communicator to use.
    """
    __slots__ = ("status","data","comm")
    def __init__(self, comm):
        import mpi4py.MPI
        self.comm = comm
        self.status = mpi4py.MPI.Status()
        self.data = None
    def probe(self, **kwds):
        """Perform a blocking test for a message"""
        self.comm.Probe(status=self.status)
        self.data = None
    def recv(self, datatype=None, data=None):
        """Complete the receive for the most recent message
        probe and return the data as a numeric array or a
        string, depending on the datatype keyword.

        Parameters
        ----------
        datatype : {``mpi4py.MPI.DOUBLE``, ``mpi4py.MPI.CHAR``}, optional
            An MPI datatype used to interpret the received
            data. If None, ``mpi4py.MPI.DOUBLE`` will be
            used. (default: None)
        data : array.array or None, optional
            An existing data array to store data into. If
            None, one will be created. (default: None)
        """
        assert not self.status.Get_error()
        if datatype is None:
            count = self.status.Get_count()
        else:
            count = self.status.Get_count(datatype=datatype)
        if count == 0:
            recv_nothing(self.comm,
                         self.status)
        else:
            self.data = recv_data(self.comm,
                                  self.status,
                                  datatype=datatype,
                                  out=data)
    @property
    def tag(self): return self.status.Get_tag()
    @property
    def source(self): return self.status.Get_source()

def recv_nothing(comm, status):
    """A helper function for receiving an empty
    message. This function is not thread safe.

    Parameters
    ----------
    comm : :class:`mpi4py.MPI.Comm`
        An MPI communicator.
    status : :class:`mpi4py.MPI.Status`
        An MPI status object that has been populated with
        information about the message to be received via a
        probe. If None, a new status object will be created
        and an empty message will be expected from any
        source with any tag. (default: None)

    Returns
    -------
    status : :class:`mpi4py.MPI.Status`
        If the original status argument was not None, it
        will be returned after being updated by the
        receive. Otherwise, the status object that was
        created will be returned.
    """
    import mpi4py.MPI
    if recv_nothing._nothing is None:
        recv_nothing._nothing = [array.array("B",[]),
                                 mpi4py.MPI.CHAR]
    assert not status.Get_error()
    assert status.Get_count(mpi4py.MPI.CHAR) == 0
    comm.Recv(recv_nothing._nothing,
              source=status.Get_source(),
              tag=status.Get_tag(),
              status=status)
    assert not status.Get_error()
    assert status.Get_count(mpi4py.MPI.CHAR) == 0
    return status
recv_nothing._nothing = None

def send_nothing(comm, dest, tag=0):
    """A helper function for sending an empty message
    with a given tag. This function is not thread safe.

    Parameters
    ----------
    comm : :class:`mpi4py.MPI.Comm`
        An MPI communicator.
    dest : int
        The process rank to send the message to.
    tag : int, optional
        A valid MPI tag to use for the message. (default: 0)
    """
    import mpi4py.MPI
    if send_nothing._nothing is None:
        send_nothing._nothing = [array.array("B",[]),
                                 mpi4py.MPI.CHAR]
    comm.Send(send_nothing._nothing,
         dest,
         tag=tag)
send_nothing._nothing = None

def recv_data(comm, status, datatype, out=None):
    """A helper function for receiving numeric or string
    data sent using the lower-level buffer-based mpi4py
    routines.

    Parameters
    ----------
    comm : :class:`mpi4py.MPI.Comm`
        An MPI communicator.
    status : :class:`mpi4py.MPI.Status`
        An MPI status object that has been populated with
        information about the message to be received via a
        probe.
    datatype : :class:`mpi4py.MPI.Datatype`
        An MPI datatype used to interpret the received
        data. If the datatype is :obj:`mpi4py.MPI.CHAR`,
        the received data will be converted to a string.
    out : buffer-like object, optional
        A buffer-like object that is compatible with the datatype
        argument and can be passed to comm.Recv. Can only be left
        as None when the datatype is :obj:`mpi4py.MPI.CHAR`.

    Returns
    -------
    string or user-provided data buffer
    """
    import mpi4py.MPI
    assert not status.Get_error()
    size = status.Get_count(datatype)
    if datatype == mpi4py.MPI.CHAR:
        assert out is None
        out = array.array("B",b"\0")*size
    assert (out is not None) and \
        (len(out) >= size)
    comm.Recv([out,datatype],
              source=status.Get_source(),
              tag=status.Get_tag(),
              status=status)
    assert not status.Get_error()
    if datatype == mpi4py.MPI.CHAR:
        out = _array_to_string(out)
    return out



"""
Branch-and-bound problem definition.

Copyright by Gabriel A. Hackebeil (gabe.hackebeil@gmail.com).
"""
import array

from pybnb.common import (minimize,
                          maximize,
                          inf)

class Problem(object):
    """The abstract base class used for defining
    branch-and-bound problems."""
    __slots__ = ()

    def infeasible_objective(self):
        """Returns the value that represents an infeasible
        objective (i.e., +inf or -inf depending on the
        sense). The :class:`Problem <pybnb.problem.Problem>`
        base class implements this method. """
        if self.sense() == minimize:
            return inf
        else:
            assert self.sense() == maximize
            return -inf

    def unbounded_objective(self):
        """Returns the value that represents an unbounded
        objective (i.e., +inf or -inf depending on the
        sense). The :class:`Problem <pybnb.problem.Problem>`
        base class implements this method."""
        if self.sense() == minimize:
            return -inf
        else:
            assert self.sense() == maximize
            return inf

    #
    # Abstract Methods
    #

    def sense(self):                              #pragma:nocover
        """Returns the objective sense for this problem.

        Note
        ----
        This method is abstract and must be defined by the
        user.
        """
        raise NotImplementedError()

    def objective(self):                          #pragma:nocover
        """Returns a feasible value for the objective of the
        current problem state or
        :func:`self.infeasible_objective()
        <pybnb.problem.Problem.infeasible_objective>` if the
        current state is not feasible.

        Note
        ----
        This method is abstract and must be defined by the
        user.
        """
        raise NotImplementedError()

    def bound(self):                              #pragma:nocover
        """Returns a value that is a bound on the objective
        of the current problem state or
        :func:`self.unbounded_objective()
        <pybnb.problem.Problem.unbounded_objective>` if no
        non-trivial bound is available.

        Note
        ----
        This method is abstract and must be defined by the
        user.
        """
        raise NotImplementedError()

    def branch(self):                #pragma:nocover
        """Returns a list of :class:`Node <pybnb.node.Node>`
        objects that partition the node state into zero or
        more children. This method can also be defined as a
        generator.

        Note
        ----
        This method is abstract and must be defined by the
        user.
        """
        raise NotImplementedError()

    def save_state(self, node):                   #pragma:nocover
        """Saves the current problem state into the given
        :class:`pybnb.node.Node` object.

        This method is guaranteed to be called once at the
        start of the solve by all processes involved to
        collect the root node problem state, but it may be
        called additional times. When it is called for the
        root node, the :attr:`node.tree_depth <pybnb.node.Node.tree_depth>`
        will be zero.

        Note
        ----
        This method is abstract and must be defined by the
        user.
        """
        raise NotImplementedError()

    def load_state(self, node):                   #pragma:nocover
        """Loads the problem state that is stored on the
        given :class:`pybnb.node.Node` object.

        Note
        ----
        This method is abstract and must be defined by the
        user.
        """
        raise NotImplementedError()

    #
    # Optional Abstract Methods
    #

    def notify_solve_begins(self,
                            comm,
                            worker_comm,
                            convergence_checker):
        """Called when a branch-and-bound solver
        begins as solve. The :class:`Problem
        <pybnb.problem.Problem>` base class provides a
        default implementation for this method that does
        nothing.

        Parameters
        ----------
        comm : ``mpi4py.MPI.Comm`` or ``None``
            The full MPI communicator that includes all
            processes. Will be None if MPI has been
            disabled.
        worker_comm : ``mpi4py.MPI.Comm`` or ``None``
            The MPI communicator that includes only worker
            processes. Will be None if MPI has been
            disabled.
        convergence_checker : :class:`ConvergenceChecker <pybnb.convergence_checker.ConvergenceChecker>`:
            The class used for comparing the objective and
            bound values during the solve.
        """
        pass

    def notify_new_best_node(self,
                             node,
                             current):
        """Called when a branch-and-bound solver receives a
        new best node from the dispatcher. The
        :class:`Problem <pybnb.problem.Problem>` base class
        provides a default implementation for this method
        that does nothing.

        Parameters
        ----------
        node : :class:`Node <pybnb.node.Node>`
            The new best node.
        current : bool
            Indicates whether or not the node argument is
            the currently loaded node (from the most recent
            :func:`load_state <pybnb.problem.load_state>`
            call).
        """
        pass

    def notify_solve_finished(self,
                              comm,
                              worker_comm,
                              results):
        """Called when a branch-and-bound solver
        finishes. The :class:`Problem
        <pybnb.problem.Problem>` base class provides a
        default implementation for this method that does
        nothing.

        Parameters
        ----------
        comm : ``mpi4py.MPI.Comm`` or ``None``
            The full MPI communicator that includes all
            processes. Will be None if MPI has been
            disabled.
        worker_comm : ``mpi4py.MPI.Comm`` or ``None``
            The MPI communicator that includes only worker
            processes. Will be None if MPI has been
            disabled.
        results : :class:`SolverResults <pybnb.solver_results.SolverResults>`
            The fully populated results container that will
            be returned from the solver.
        """
        pass

class _SolveInfo(object):
    __slots__ = ("data")
    _data_size = 11
    def __init__(self):
        self.data = array.array('d',[0]) * \
                    _SolveInfo._data_size

    def reset(self):
        """Resets all statistics to zero."""
        for i in range(len(self.data)):
            self.data[i] = 0.0

    def add_from(self, other):
        if type(other) is not _SolveInfo:
            raise TypeError("Type %s can not be added "
                            "with a _SolveInfo object"
                            % (other.__class__.__name__))
        assert len(self.data) == len(other.data)
        for i in range(_SolveInfo._data_size):
            self.data[i] += other.data[i]

    @property
    def total_queue_time(self):
        return float(self.data[0])
    @total_queue_time.setter
    def total_queue_time(self, val):
        self.data[0] = val

    @property
    def queue_call_count(self):
        return int(self.data[1])
    @queue_call_count.setter
    def queue_call_count(self, val):
        self.data[1] = val

    def _increment_queue_stat(self, time_, count):
        self.data[0] += time_
        self.data[1] += count

    @property
    def total_objective_time(self):
        return float(self.data[2])
    @total_objective_time.setter
    def total_objective_time(self, val):
        self.data[2] = val

    @property
    def objective_call_count(self):
        return int(self.data[3])
    @objective_call_count.setter
    def objective_call_count(self, val):
        self.data[3] = val

    def _increment_objective_stat(self, time_, count):
        self.data[2] += time_
        self.data[3] += count

    @property
    def total_bound_time(self):
        return float(self.data[4])
    @total_bound_time.setter
    def total_bound_time(self, val):
        self.data[4] = val

    @property
    def bound_call_count(self):
        return int(self.data[5])
    @bound_call_count.setter
    def bound_call_count(self, val):
        self.data[5] = val

    def _increment_bound_stat(self, time_, count):
        self.data[4] += time_
        self.data[5] += count

    @property
    def total_branch_time(self):
        return float(self.data[6])
    @total_branch_time.setter
    def total_branch_time(self, val):
        self.data[6] = val

    @property
    def branch_call_count(self):
        return int(self.data[7])
    @branch_call_count.setter
    def branch_call_count(self, val):
        self.data[7] = val

    def _increment_branch_stat(self, time_, count):
        self.data[6] += time_
        self.data[7] += count

    @property
    def total_load_state_time(self):
        return float(self.data[8])
    @total_load_state_time.setter
    def total_load_state_time(self, val):
        self.data[8] = val

    @property
    def load_state_call_count(self):
        return int(self.data[9])
    @load_state_call_count.setter
    def load_state_call_count(self, val):
        self.data[9] = val

    def _increment_load_state_stat(self, time_, count):
        self.data[8] += time_
        self.data[9] += count

    @property
    def explored_nodes_count(self):
        return int(self.data[10])
    @explored_nodes_count.setter
    def explored_nodes_count(self, val):
        self.data[10] = val

    def _increment_explored_nodes_stat(self, count):
        self.data[10] += count


class _ProblemWithSolveInfoCollection(Problem):
    """A Problem objects that keeps track of statistics used
    by the solver"""
    def __init__(self):
        self._clock = None
        self._solve_info = None

    def set_clock(self, clock):
        self._clock = clock

    def set_solve_info_object(self, solve_info):
        assert isinstance(solve_info, _SolveInfo)
        self._solve_info = solve_info

class _SimpleSolveInfoCollector(_ProblemWithSolveInfoCollection):
    """A wrapper for Problem objects that collects statistics
    on methods called during the solve."""
    def __init__(self, problem):
        self._problem = problem
        super(_SimpleSolveInfoCollector, self).__init__()

    #
    # Define Problem methods
    #

    def sense(self):
        return self._problem.sense()

    def objective(self):
        start = self._clock()
        tmp = self._problem.objective()
        stop = self._clock()
        self._solve_info._increment_objective_stat(stop-start, 1)
        return tmp

    def bound(self):
        start = self._clock()
        tmp = self._problem.bound()
        stop = self._clock()
        self._solve_info._increment_bound_stat(stop-start, 1)
        return tmp

    def branch(self):
        start = self._clock()
        for item in self._problem.branch():
            yield item
        stop = self._clock()
        self._solve_info._increment_branch_stat(stop-start, 1)

    def save_state(self, node):
        self._problem.save_state(node)

    def load_state(self, node):
        start = self._clock()
        self._problem.load_state(node)
        stop = self._clock()
        self._solve_info._increment_load_state_stat(stop-start, 1)

    def notify_solve_begins(self,
                            comm,
                            worker_comm,
                            convergence_checker):
        self._problem.notify_solve_begins(
            comm,
            worker_comm,
            convergence_checker)

    def notify_new_best_node(self,
                             node,
                             current):
        self._problem.notify_new_best_node(node,
                                           current)

    def notify_solve_finished(self,
                              comm,
                              worker_comm,
                              results):
        self._problem.notify_solve_finished(
            comm,
            worker_comm,
            results)



import random

import pytest

from pybnb.common import (minimize,
                          maximize,
                          inf)
from pybnb.node import Node
from pybnb.priority_queue import \
    (_NoThreadingMaxPriorityFirstQueue,
     _NoThreadingFIFOQueue,
     _NoThreadingLIFOQueue,
     WorstBoundFirstPriorityQueue,
     CustomPriorityQueue,
     BestObjectiveFirstPriorityQueue,
     BreadthFirstPriorityQueue,
     DepthFirstPriorityQueue,
     FIFOQueue,
     LIFOQueue,
     RandomPriorityQueue,
     LocalGapPriorityQueue,
     LexicographicPriorityQueue,
     PriorityQueueFactory,
     register_queue_type)

def _new_child(node):
    child = Node()
    child.objective = node.objective
    child.bound = node.bound
    child.tree_depth = node.tree_depth + 1
    assert child.queue_priority is None
    assert child.state is None
    return child

def assert_isheap(x):
    for k in range(len(x)):
        if ((2*k) + 1) < len(x):
            assert x[k] <= x[2*k+1]
        if ((2*k) + 2) < len(x):
            assert x[k] <= x[2*k+2]

class TestFactory(object):

    def test_factory(self):
        assert type(PriorityQueueFactory('bound', minimize)) is \
            WorstBoundFirstPriorityQueue
        assert type(PriorityQueueFactory('custom', minimize)) is \
            CustomPriorityQueue
        assert type(PriorityQueueFactory('objective', minimize)) is \
            BestObjectiveFirstPriorityQueue
        assert type(PriorityQueueFactory('breadth', minimize)) is \
            BreadthFirstPriorityQueue
        assert type(PriorityQueueFactory('depth', minimize)) is \
            DepthFirstPriorityQueue
        assert type(PriorityQueueFactory('fifo', minimize)) is \
            FIFOQueue
        assert type(PriorityQueueFactory('lifo', minimize)) is \
            LIFOQueue
        assert type(PriorityQueueFactory('random', minimize)) is \
            RandomPriorityQueue
        assert type(PriorityQueueFactory('local_gap', minimize)) is \
            LocalGapPriorityQueue
        with pytest.raises(ValueError):
            PriorityQueueFactory('_not_a_type_', minimize)
        # test LexicographicPriorityQueue creation
        assert type(PriorityQueueFactory(('bound','objective'), minimize)) is \
            LexicographicPriorityQueue
        with pytest.raises(ValueError):
            PriorityQueueFactory(('_not_a_type_',), minimize)
        with pytest.raises(ValueError):
            PriorityQueueFactory(('custom',), minimize)
        with pytest.raises(ValueError):
            PriorityQueueFactory(('fifo','custom'), minimize)
        with pytest.raises(ValueError):
            PriorityQueueFactory(('custom','fifo'), minimize)
        with pytest.raises(ValueError):
            PriorityQueueFactory((), minimize)

    def test_register_queue_type(self):
        assert PriorityQueueFactory._types['bound'] is \
            WorstBoundFirstPriorityQueue
        # its okay to re-register the exact same thing
        register_queue_type('bound',
                            WorstBoundFirstPriorityQueue)
        assert PriorityQueueFactory._types['bound'] is \
            WorstBoundFirstPriorityQueue
        with pytest.raises(ValueError):
            register_queue_type('bound', None)
        assert PriorityQueueFactory._types['bound'] is \
            WorstBoundFirstPriorityQueue
        assert '_not_a_type_' not in PriorityQueueFactory._types
        try:
            register_queue_type('_not_a_type_', None)
            assert PriorityQueueFactory._types['_not_a_type_'] is None
        finally:
            PriorityQueueFactory._types.pop('_not_a_type_',None)

class Test_NoThreadingMaxPriorityFirstQueue(object):

    def test_size(self):
        q = _NoThreadingMaxPriorityFirstQueue()
        assert_isheap(q._heap)
        assert q.size() == 0
        with pytest.raises(IndexError):
            q.next()
        cntr = {}
        for i in range(10):
            cntr[i] = q.put(i,0)
            assert q.size() == i+1
            assert_isheap(q._heap)
        for i in range(10):
            c_, x_ = q.next()
            assert c_ == cntr[x_]
            x = q.get()
            assert x_ == x
            assert x == i
            assert q.size() == 10-(i+1)
            assert_isheap(q._heap)
        assert q.size() == 0
        assert q.get() is None
        with pytest.raises(IndexError):
            q.next()
        assert_isheap(q._heap)

    def test_put_get(self):
        q = _NoThreadingMaxPriorityFirstQueue()
        with pytest.raises(ValueError):
            q.put(None, 0)
        with pytest.raises(ValueError):
            q.put_get(None, 0)
        assert sorted(q.items()) == []
        assert_isheap(q._heap)
        assert q.put_get(1,1) == (0,1)
        assert q.put_get(1,1) == (1,1)
        cntr = {}
        cntr[1] = q.put(1,1)
        assert sorted(q.items()) == [1]
        assert_isheap(q._heap)
        cntr[2] = q.put(2,2)
        assert sorted(q.items()) == [1,2]
        assert_isheap(q._heap)
        cntr[3] = q.put(3,2)
        assert sorted(q.items()) == [1,2,3]
        assert_isheap(q._heap)
        cntr[4], item_ = q.put_get(4,-4)
        assert item_ == 2
        assert sorted(q.items()) == [1,3,4]
        assert_isheap(q._heap)
        assert q.next() == (cntr[q.next()[1]], 3)
        assert q.get() == 3
        assert sorted(q.items()) == [1,4]
        assert_isheap(q._heap)
        assert q.next() == (cntr[q.next()[1]], 1)
        assert q.get() == 1
        assert sorted(q.items()) == [4]
        assert_isheap(q._heap)
        assert q.next() == (cntr[q.next()[1]], 4)
        assert q.get() == 4
        assert sorted(q.items()) == []
        assert_isheap(q._heap)
        with pytest.raises(IndexError):
            q.next()

    def test_filter(self):
        for k in range(-10,11):
            cutoff = k*11
            def _filter(item):
                if item <= cutoff:
                    return True
            items = list(range(-1000,1000))
            random.shuffle(items)
            q = _NoThreadingMaxPriorityFirstQueue()
            for i in items:
                q.put(i, 1)
            assert_isheap(q._heap)
            correct = []
            removed = []
            for _,_,item in q._heap:
                if item <= cutoff:
                    correct.append(item)
                else:
                    removed.append(item)
            removed_ = q.filter(_filter)
            assert removed_ == removed
            assert_isheap(q._heap)
            check = []
            for _,_,item in q._heap:
                check.append(item)
            assert sorted(correct) == sorted(check)

class Test_NoThreadingFIFOQueue(object):

    def test_size(self):
        q = _NoThreadingFIFOQueue()
        assert q.size() == 0
        with pytest.raises(IndexError):
            q.next()
        cntr = {}
        for i in range(10):
            cntr[i] = q.put(i)
            assert q.size() == i+1
        for i in range(10):
            c_, x_ = q.next()
            assert c_ == cntr[x_]
            x = q.get()
            assert x_ == x
            assert x == i
            assert q.size() == 10-(i+1)
        assert q.size() == 0
        assert q.get() is None
        with pytest.raises(IndexError):
            q.next()

    def test_put_get(self):
        q = _NoThreadingFIFOQueue()
        with pytest.raises(ValueError):
            q.put(None)
        with pytest.raises(ValueError):
            q.put_get(None)
        assert list(q.items()) == []
        assert q.put_get(1) == (0,1)
        assert q.put_get(1) == (1,1)
        cntr = {}
        cntr[1] = q.put(1)
        assert list(q.items()) == [1]
        cntr[2] = q.put(2)
        assert list(q.items()) == [1,2]
        cntr[3] = q.put(3)
        assert list(q.items()) == [1,2,3]
        cntr[4], item_ = q.put_get(4)
        assert item_ == 1
        assert list(q.items()) == [2,3,4]
        assert q.next() == (cntr[q.next()[1]], 2)
        assert q.get() == 2
        assert list(q.items()) == [3,4]
        assert q.next() == (cntr[q.next()[1]], 3)
        assert q.get() == 3
        assert list(q.items()) == [4]
        assert q.next() == (cntr[q.next()[1]], 4)
        assert q.get() == 4
        assert list(q.items()) == []
        with pytest.raises(IndexError):
            q.next()

    def test_filter(self):
        for k in range(-10,11):
            cutoff = k*11
            def _filter(item):
                if item <= cutoff:
                    return True
            items = list(range(-1000,1000))
            random.shuffle(items)
            q = _NoThreadingFIFOQueue()
            for i in items:
                q.put(i)
            correct = []
            removed = []
            for _,item in q._deque:
                if item <= cutoff:
                    correct.append(item)
                else:
                    removed.append(item)
            removed_ = q.filter(_filter)
            assert removed_ == removed
            check = []
            for _,item in q._deque:
                check.append(item)
            assert sorted(correct) == sorted(check)
        for k in range(-10,11):
            cutoff = k*11
            def _filter(item):
                if item <= cutoff:
                    return True
            items = list(range(-1000,1000))
            random.shuffle(items)
            q = _NoThreadingFIFOQueue()
            for i in items:
                q.put(i)
            correct = []
            removed = []
            for cnt,item in q._deque:
                if item <= cutoff:
                    correct.append(item)
                else:
                    removed.append((cnt,item))
            removed_ = q.filter(_filter, include_counters=True)
            assert removed_ == removed
            check = []
            for _,item in q._deque:
                check.append(item)
            assert sorted(correct) == sorted(check)

class Test_NoThreadingLIFOQueue(object):

    def test_size(self):
        q = _NoThreadingLIFOQueue()
        assert q.size() == 0
        with pytest.raises(IndexError):
            q.next()
        cntr = {}
        for i in range(10):
            cntr[i] = q.put(i)
            assert q.size() == i+1
        for i in reversed(range(10)):
            c_, x_ = q.next()
            assert c_ == cntr[x_]
            x = q.get()
            assert x_ == x
            assert x == i
            assert q.size() == i
        assert q.size() == 0
        assert q.get() is None
        with pytest.raises(IndexError):
            q.next()

    def test_put_get(self):
        q = _NoThreadingLIFOQueue()
        with pytest.raises(ValueError):
            q.put(None)
        with pytest.raises(ValueError):
            q.put_get(None)
        assert list(q.items()) == []
        assert q.put_get(1) == (0,1)
        assert q.put_get(1) == (1,1)
        cntr = {}
        cntr[1] = q.put(1)
        assert list(q.items()) == [1]
        cntr[2] = q.put(2)
        assert list(q.items()) == [1,2]
        cntr[3] = q.put(3)
        assert list(q.items()) == [1,2,3]
        cntr[4], item_ = q.put_get(4)
        assert item_ == 4
        assert list(q.items()) == [1,2,3]
        assert q.next() == (cntr[q.next()[1]], 3)
        assert q.get() == 3
        assert list(q.items()) == [1,2]
        assert q.next() == (cntr[q.next()[1]], 2)
        assert q.get() == 2
        assert list(q.items()) == [1]
        assert q.next() == (cntr[q.next()[1]], 1)
        assert q.get() == 1
        assert list(q.items()) == []
        with pytest.raises(IndexError):
            q.next()

    def test_filter(self):
        for k in range(-10,11):
            cutoff = k*11
            def _filter(item):
                if item <= cutoff:
                    return True
            items = list(range(-1000,1000))
            random.shuffle(items)
            q = _NoThreadingLIFOQueue()
            for i in items:
                q.put(i)
            correct = []
            removed = []
            for _,item in q._items:
                if item <= cutoff:
                    correct.append(item)
                else:
                    removed.append(item)
            removed_ = q.filter(_filter)
            assert removed_ == removed
            check = []
            for _,item in q._items:
                check.append(item)
            assert sorted(correct) == sorted(check)
        for k in range(-10,11):
            cutoff = k*11
            def _filter(item):
                if item <= cutoff:
                    return True
            items = list(range(-1000,1000))
            random.shuffle(items)
            q = _NoThreadingLIFOQueue()
            for i in items:
                q.put(i)
            correct = []
            removed = []
            for cnt,item in q._items:
                if item <= cutoff:
                    correct.append(item)
                else:
                    removed.append((cnt,item))
            removed_ = q.filter(_filter, include_counters=True)
            assert removed_ == removed
            check = []
            for _,item in q._items:
                check.append(item)
            assert sorted(correct) == sorted(check)

def _check_items(q, items):
    found = dict((id(item), False)
                 for item in items)
    assert len(list(q.items())) == len(items)
    for queue_item in q.items():
        assert id(queue_item) in found
        assert not found[id(queue_item)]
        found[id(queue_item)] = True
    assert all(found.values())

class TestWorstBoundFirstPriorityQueue(object):

    def test_overwrites_queue_priority(self):
        q = WorstBoundFirstPriorityQueue(minimize)
        node = Node()
        node.bound = 1
        assert node.queue_priority is None
        q.put(node)
        assert node.queue_priority == -1

        q = WorstBoundFirstPriorityQueue(maximize)
        node = Node()
        node.bound = 1
        assert node.queue_priority is None
        q.put(node)
        assert node.queue_priority == 1

    def test_usage_minimize(self):
        q = WorstBoundFirstPriorityQueue(minimize)
        assert q.size() == 0
        assert q.bound() is None
        assert len(list(q.items())) == 0
        assert q.get() is None
        items = []
        for i in range(1,11):
            node = Node()
            node.bound = i
            assert node.queue_priority is None
            cnt_ = q.put(node)
            assert cnt_ == i-1
            assert node.queue_priority == -node.bound
            items.append(node)
            assert q.size() == i
            assert q.bound() == 1
            _check_items(q, items)
        assert q.size() == 10
        assert q.bound() == 1
        removed = q.filter(lambda n_: n_.bound >= 5)
        assert q.size() == 6
        assert len(removed) == 4
        for node_ in removed:
            assert node_.bound < 5
        assert q.bound() == 5
        for i in range(5, 11):
            node = q.get()
            assert node.bound == i
            assert node.queue_priority == -node.bound
            if i != 10:
                assert q.bound() == i+1
            else:
                assert q.bound() is None
        assert q.size() == 0
        node = Node()
        node.bound = -1
        assert node.queue_priority is None
        cnt_ = q.put(node)
        node_ = q.get()
        assert cnt_ == 10
        assert node_ is node
        assert node.queue_priority == 1
        node.bound = -2
        cnt_ = q.put(node)
        assert node.queue_priority == 2
        assert cnt_ == 11
        node2 = Node()
        node2.bound = -3
        assert node2.queue_priority is None
        cnt_ = q.put(node2)
        node_ = q.get()
        assert cnt_ == 12
        assert node_ is node2
        assert node2.queue_priority == 3
        node2.bound = -1
        cnt_ = q.put(node2)
        node_ = q.get()
        assert node2.queue_priority == 1
        assert cnt_ == 13
        assert node_ is node
        assert q.size() == 1

    def test_usage_maximize(self):
        q = WorstBoundFirstPriorityQueue(maximize)
        assert q.size() == 0
        assert q.bound() is None
        assert len(list(q.items())) == 0
        assert q.get() is None
        items = []
        for i in range(1,11):
            node = Node()
            node.bound = -i
            assert node.queue_priority is None
            cnt_ = q.put(node)
            assert cnt_ == i-1
            assert node.queue_priority == node.bound
            items.append(node)
            assert q.size() == i
            assert q.bound() == -1
            _check_items(q, items)
        assert q.size() == 10
        assert q.bound() == -1
        removed = q.filter(lambda n_: n_.bound <= -5)
        assert q.size() == 6
        assert len(removed) == 4
        for node_ in removed:
            assert node_.bound > -5
        assert q.bound() == -5
        for i in range(5, 11):
            node = q.get()
            assert node.bound == -i
            assert node.queue_priority == node.bound
            if i != 10:
                assert q.bound() == -i-1
            else:
                assert q.bound() is None
        assert q.size() == 0
        node = Node()
        node.bound = 1
        assert node.queue_priority is None
        cnt_ = q.put(node)
        node_ = q.get()
        assert cnt_ == 10
        assert node_ is node
        assert node.queue_priority == 1
        node.bound = 2
        cnt_ = q.put(node)
        assert node.queue_priority == 2
        assert cnt_ == 11
        node2 = Node()
        node2.bound = 3
        assert node2.queue_priority is None
        cnt_ = q.put(node2)
        node_ = q.get()
        assert cnt_ == 12
        assert node_ is node2
        assert node2.queue_priority == 3
        node2.bound = 1
        cnt_ = q.put(node2)
        node_ = q.get()
        assert node2.queue_priority == 1
        assert cnt_ == 13
        assert node_ is node
        assert q.size() == 1

class TestCustomPriorityQueue(object):

    def test_missing_queue_priority(self):
        q = CustomPriorityQueue(minimize)
        node = Node()
        node.tree_depth = 0
        node.bound = 0
        assert node.queue_priority is None
        with pytest.raises(ValueError):
            q.put(node)
        node.queue_priority = 1
        q.put(node)
        child = _new_child(node)
        assert child.queue_priority is None
        with pytest.raises(ValueError):
            q.put(child)

    def test_usage_minimize(self):
        q = CustomPriorityQueue(minimize)
        assert q.size() == 0
        assert q.bound() is None
        assert len(list(q.items())) == 0
        assert q.get() is None
        items = []
        for i in range(1,11):
            node = Node()
            node.bound = i
            node.queue_priority = -i
            assert q.put(node) == i-1
            assert node.queue_priority == -i
            items.append(node)
            assert q.size() == i
            assert q.bound() == 1
            _check_items(q, items)
        assert q.size() == 10
        assert q.bound() == 1
        removed = q.filter(lambda n_: n_.bound >= 5)
        assert q.size() == 6
        assert len(removed) == 4
        for node_ in removed:
            assert node_.bound < 5
        assert q.bound() == 5
        for i in range(5, 11):
            node = q.get()
            assert node.bound == i
            assert node.queue_priority == -i
            if i != 10:
                assert q.bound() == i+1
            else:
                assert q.bound() is None
        assert q.size() == 0
        node = Node()
        node.bound = 0
        node.queue_priority = 1
        cnt_ = q.put(node)
        node_ = q.get()
        assert cnt_ == 10
        assert node_ is node
        assert node.queue_priority == 1
        assert q.bound() is None
        node.queue_priority = 2
        cnt_ = q.put(node)
        assert node.queue_priority == 2
        assert cnt_ == 11
        assert q.bound() == 0
        node2 = Node()
        node2.bound = 1
        node2.queue_priority = 3
        cnt_ = q.put(node2)
        node_ = q.get()
        assert cnt_ == 12
        assert node_ is node2
        assert node2.queue_priority == 3
        assert q.bound() == 0
        node2.queue_priority = 1
        cnt_ = q.put(node2)
        node_ = q.get()
        assert node2.queue_priority == 1
        assert cnt_ == 13
        assert node_ is node
        assert q.size() == 1
        assert q.bound() == 1

    def test_usage_maximize(self):
        q = CustomPriorityQueue(maximize)
        assert q.size() == 0
        assert q.bound() is None
        assert len(list(q.items())) == 0
        assert q.get() is None
        items = []
        for i in range(1,11):
            node = Node()
            node.bound = -i
            node.queue_priority = i
            assert q.put(node) == i-1
            assert node.queue_priority == i
            items.append(node)
            assert q.size() == i
            assert q.bound() == -1
            _check_items(q, items)
        assert q.size() == 10
        assert q.bound() == -1
        removed = q.filter(lambda n_: n_.bound <= -5)
        assert q.size() == 6
        assert len(removed) == 4
        for node_ in removed:
            assert node_.bound > -5
        assert q.bound() == -5
        for i in range(10, 4, -1):
            node = q.get()
            assert node.bound == -i
            assert node.queue_priority == i
            if i != 5:
                assert q.bound() == -5
            else:
                assert q.bound() is None
        assert q.size() == 0
        node = Node()
        node.bound = 0
        node.queue_priority = 1
        cnt_ = q.put(node)
        node_ = q.get()
        assert cnt_ == 10
        assert node_ is node
        assert node.queue_priority == 1
        assert q.bound() is None
        node.queue_priority = 2
        cnt_ = q.put(node)
        assert node.queue_priority == 2
        assert cnt_ == 11
        assert q.bound() == 0
        node2 = Node()
        node2.bound = 1
        node2.queue_priority = 3
        cnt_ = q.put(node2)
        node_ = q.get()
        assert cnt_ == 12
        assert node_ is node2
        assert node2.queue_priority == 3
        assert q.bound() == 0
        node2.queue_priority = 1
        cnt_ = q.put(node2)
        node_ = q.get()
        assert node2.queue_priority == 1
        assert cnt_ == 13
        assert node_ is node
        assert q.size() == 1
        assert q.bound() == 1

class TestBestObjectiveFirstPriorityQueue(object):

    def test_overwrites_queue_priority(self):
        q = BestObjectiveFirstPriorityQueue(minimize)
        node = Node()
        node.tree_depth = 0
        node.bound = -1
        assert node.queue_priority is None
        node.objective = 1
        assert q.put(node) == 0
        assert node.objective == 1
        assert node.queue_priority == -1
        child = _new_child(node)
        assert child.objective == 1
        child.objective = 0
        assert child.queue_priority is None
        cnt = q.put(child)
        node_ = q.get()
        assert child.queue_priority == 0
        assert cnt == 1
        assert node_ is child
        child.objective = 2
        cnt = q.put(child)
        node_ = q.get()
        assert child.queue_priority == -2
        assert cnt == 2
        assert node_ is node
        assert q.bound() == -1

        q = BestObjectiveFirstPriorityQueue(maximize)
        node = Node()
        node.tree_depth = 0
        node.bound = 3
        assert node.queue_priority is None
        node.objective = 1
        assert q.put(node) == 0
        assert node.objective == 1
        assert node.queue_priority == 1
        child = _new_child(node)
        assert child.objective == 1
        child.objective = 2
        assert child.queue_priority is None
        cnt = q.put(child)
        node_ = q.get()
        assert child.queue_priority == 2
        assert cnt == 1
        assert node_ is child
        child.objective = 0
        cnt = q.put(child)
        node_ = q.get()
        assert child.queue_priority == 0
        assert cnt == 2
        assert node_ is node
        assert q.bound() == 3

class TestBreadthFirstPriorityQueue(object):

    def test_overwrites_queue_priority(self):
        q = BreadthFirstPriorityQueue(minimize)
        node = Node()
        node.tree_depth = 0
        node.bound = 0
        assert node.queue_priority is None
        assert q.put(node) == 0
        assert node.tree_depth == 0
        assert node.queue_priority == 0
        child = _new_child(node)
        assert child.tree_depth == 1
        assert child.queue_priority is None
        assert q.put(child) == 1
        assert child.queue_priority == -child.tree_depth

        l1 = Node()
        l1.tree_depth = 0
        l1.bound = 1
        l2 = _new_child(l1)
        l3 = _new_child(l2)
        q = BreadthFirstPriorityQueue(minimize)
        q.put(l2)
        cnt = q.put(l1)
        node_ = q.get()
        assert cnt == 1
        assert node_ is l1
        cnt = q.put(l3)
        node_ = q.get()
        assert cnt == 2
        assert node_ is l2
        assert q.bound() == 1

class TestDepthFirstPriorityQueue(object):

    def test_overwrites_queue_priority(self):
        q = DepthFirstPriorityQueue(minimize)
        node = Node()
        node.tree_depth = 0
        node.bound = 0
        assert node.queue_priority is None
        assert q.put(node) == 0
        assert node.tree_depth == 0
        assert node.queue_priority == 0
        child = _new_child(node)
        assert child.tree_depth == 1
        assert child.queue_priority is None
        assert q.put(child) == 1
        assert child.queue_priority == child.tree_depth

        l1 = Node()
        l1.tree_depth = 0
        l1.bound = 1
        l2 = _new_child(l1)
        l3 = _new_child(l2)
        q = DepthFirstPriorityQueue(minimize)
        q.put(l2)
        cnt = q.put(l3)
        node_ = q.get()
        assert cnt == 1
        assert node_ is l3
        cnt = q.put(l2)
        node_ = q.get()
        assert cnt == 2
        assert node_ is l2
        assert q.bound() == 1

class TestFIFOQueue(object):

    def test_overwrites_queue_priority(self):
        q = FIFOQueue(minimize)
        node = Node()
        node.tree_depth = 0
        node.bound = 0
        assert node.queue_priority is None
        assert q.put(node) == 0
        assert node.queue_priority == 0
        child = _new_child(node)
        assert child.queue_priority is None
        assert q.put(child) == 1
        assert child.queue_priority == -1

        l1 = Node()
        l1.tree_depth = 0
        l1.bound = 1
        l2 = _new_child(l1)
        l3 = _new_child(l2)
        q = FIFOQueue(minimize)
        cnt = q.put(l2)
        node_ = q.get()
        assert cnt == 0
        assert node_ is l2
        cnt = q.put(l2)
        assert cnt == 1
        cnt = q.put(l3)
        node_ = q.get()
        assert cnt == 2
        assert node_ is l2
        cnt = q.put(l1)
        node_ = q.get()
        assert cnt == 3
        assert node_ is l3
        assert q.bound() == 1

class TestLIFOQueue(object):

    def test_overwrites_queue_priority(self):
        q = LIFOQueue(minimize)
        node = Node()
        node.tree_depth = 0
        node.bound = 0
        assert node.queue_priority is None
        assert q.put(node) == 0
        assert node.queue_priority == 0
        child = _new_child(node)
        assert child.queue_priority is None
        assert q.put(child) == 1
        assert child.queue_priority == 1

        l1 = Node()
        l1.tree_depth = 0
        l1.bound = 1
        l2 = _new_child(l1)
        l3 = _new_child(l2)
        q = LIFOQueue(minimize)
        cnt = q.put(l2)
        node_ = q.get()
        assert cnt == 0
        assert node_ is l2
        cnt = q.put(l2)
        assert q.bound() == 1
        assert cnt == 1
        cnt = q.put(l3)
        node_ = q.get()
        assert cnt == 2
        assert node_ is l3
        node_ = q.get()
        assert node_ is l2
        assert q.bound() is None

class TestRandomPriorityQueue(object):

    def test_overwrites_queue_priority(self):
        q = RandomPriorityQueue(minimize)
        node = Node()
        node.tree_depth = 0
        node.bound = 0
        assert node.queue_priority is None
        assert q.put(node) == 0
        assert node.queue_priority is not None
        assert 0 <= node.queue_priority <= 1
        child = _new_child(node)
        assert child.queue_priority is None
        assert q.put(child) == 1
        assert child.queue_priority is not None
        assert 0 <= child.queue_priority <= 1

        l1 = Node()
        l1.tree_depth = 0
        l1.bound = 1
        l2 = _new_child(l1)
        l3 = _new_child(l2)
        q = RandomPriorityQueue(minimize)
        assert l2.queue_priority is None
        cnt = q.put(l2)
        node_ = q.get()
        assert node_ is l2
        assert l2.queue_priority is not None
        assert 0 <= l2.queue_priority <= 1
        assert cnt == 0
        cnt = q.put(l2)
        assert cnt == 1
        assert l3.queue_priority is None
        cnt = q.put(l3)
        node_ = q.get()
        assert cnt == 2
        assert l3.queue_priority is not None
        assert 0 <= l3.queue_priority <= 1
        assert node_ is max([l2, l3],
                            key=lambda x_: x_.queue_priority)

class TestLocalGapPriorityQueue(object):

    def test_overwrites_queue_priority(self):
        q = LocalGapPriorityQueue(minimize)
        node = Node()
        node.tree_depth = 0
        node.bound = -inf
        node.objective = inf
        assert node.queue_priority is None
        assert q.put(node) == 0
        assert node.queue_priority is not None
        assert node.queue_priority == inf
        child = _new_child(node)
        assert child.bound == -inf
        assert child.objective == inf
        child.bound = 0
        assert child.queue_priority is None
        assert q.put(child) == 1
        assert child.queue_priority is not None
        assert child.queue_priority == inf
        child = _new_child(child)
        assert child.bound == 0
        assert child.objective == inf
        child.objective = 1
        assert child.queue_priority is None
        assert q.put(child) == 2
        assert child.queue_priority is not None
        assert child.queue_priority == 1

        l1 = Node()
        l1.tree_depth = 0
        l1.bound = 1
        l1.objective = 5
        l2 = _new_child(l1)
        l3 = _new_child(l2)
        q = LocalGapPriorityQueue(minimize)
        assert l2.queue_priority is None
        cnt = q.put(l2)
        node_ = q.get()
        assert node_ is l2
        assert l2.queue_priority is not None
        assert l2.queue_priority == 4
        assert cnt == 0
        cnt = q.put(l2)
        assert cnt == 1
        assert l3.queue_priority is None
        l3.objective = 6
        cnt = q.put(l3)
        node_ = q.get()
        assert cnt == 2
        assert l3.queue_priority is not None
        assert l3.queue_priority == 5
        assert node_ is l3

        q = LocalGapPriorityQueue(maximize)
        node = Node()
        node.tree_depth = 0
        node.bound = inf
        node.objective = -inf
        assert node.queue_priority is None
        assert q.put(node) == 0
        assert node.queue_priority is not None
        assert node.queue_priority == inf
        child = _new_child(node)
        assert child.bound == inf
        assert child.objective == -inf
        child.bound = 0
        assert child.queue_priority is None
        assert q.put(child) == 1
        assert child.queue_priority is not None
        assert child.queue_priority == inf
        child = _new_child(child)
        assert child.bound == 0
        assert child.objective == -inf
        child.objective = -1
        assert child.queue_priority is None
        assert q.put(child) == 2
        assert child.queue_priority is not None
        assert child.queue_priority == 1

        l1 = Node()
        l1.tree_depth = 0
        l1.bound = -1
        l1.objective = -5
        l2 = _new_child(l1)
        l3 = _new_child(l2)
        q = LocalGapPriorityQueue(maximize)
        assert l2.queue_priority is None
        cnt = q.put(l2)
        node_ = q.get()
        assert node_ is l2
        assert l2.queue_priority is not None
        assert l2.queue_priority == 4
        assert cnt == 0
        cnt = q.put(l2)
        assert cnt == 1
        assert l3.queue_priority is None
        l3.objective = -6
        cnt = q.put(l3)
        node_ = q.get()
        assert cnt == 2
        assert l3.queue_priority is not None
        assert l3.queue_priority == 5
        assert node_ is l3

class TestLexicographicPriorityQueue(object):

    def test_overwrites_queue_priority(self):
        # min
        q = LexicographicPriorityQueue(
            (WorstBoundFirstPriorityQueue,
             BestObjectiveFirstPriorityQueue),
            minimize)
        node = Node()
        node.tree_depth = 0
        node.bound = 0
        node.objective = 2
        assert node.queue_priority is None
        assert q.put(node) == 0
        assert node.queue_priority is not None
        assert node.queue_priority == (0,-2)
        c1 = child = _new_child(node)
        assert child.bound == 0
        assert child.objective == 2
        child.objective = 1
        assert child.queue_priority is None
        assert q.put(child) == 1
        assert child.queue_priority is not None
        assert child.queue_priority == (0,-1)
        c2 = child = _new_child(child)
        assert child.bound == 0
        assert child.objective == 1
        child.bound = -1
        child.objective = 2
        assert child.queue_priority is None
        assert q.put(child) == 2
        assert child.queue_priority is not None
        assert child.queue_priority == (1,-2)
        c3 = child = _new_child(child)
        assert child.bound == -1
        assert child.objective == 2
        child.bound = 1
        child.objective = -100
        assert child.queue_priority is None
        assert q.put(child) == 3
        assert child.queue_priority is not None
        assert child.queue_priority == (-1,100)
        assert q.get() is c2
        assert q.get() is c1
        assert q.get() is node
        assert q.get() is c3
        assert q.get() is None

        # max
        q = LexicographicPriorityQueue(
            (WorstBoundFirstPriorityQueue,
             BestObjectiveFirstPriorityQueue),
            maximize)
        node = Node()
        node.tree_depth = 0
        node.bound = 0
        node.objective = -2
        assert node.queue_priority is None
        assert q.put(node) == 0
        assert node.queue_priority is not None
        assert node.queue_priority == (0,-2)
        c1 = child = _new_child(node)
        assert child.bound == 0
        assert child.objective == -2
        child.objective = -1
        assert child.queue_priority is None
        assert q.put(child) == 1
        assert child.queue_priority is not None
        assert child.queue_priority == (0,-1)
        c2 = child = _new_child(child)
        assert child.bound == 0
        assert child.objective == -1
        child.bound = 1
        child.objective = -2
        assert child.queue_priority is None
        assert q.put(child) == 2
        assert child.queue_priority is not None
        assert child.queue_priority == (1,-2)
        c3 = child = _new_child(child)
        assert child.bound == 1
        assert child.objective == -2
        child.bound = -1
        child.objective = 100
        assert child.queue_priority is None
        assert q.put(child) == 3
        assert child.queue_priority is not None
        assert child.queue_priority == (-1,100)
        assert q.get() is c2
        assert q.get() is c1
        assert q.get() is node
        assert q.get() is c3
        assert q.get() is None



import array

from pybnb.pyomo.misc import mpi_partition

from ..common import mpi_available
from runtests.mpi import MPITest

def _test_mpi_partition(comm):
    test_ranks = [0]
    if comm is not None:
        import mpi4py.MPI
        test_ranks = list(range(comm.size))
    for x in ([],
              ['a'],
              ['a','b'],
              ['a','b','c'],
              ['a','b','c']*2,
              ['a','b','c']*4,
              ['a','b','c']*16,
              ['a','b','c']*32):
        for root in test_ranks:
            x_accessed_local = array.array('i',[0])*len(x)
            for i, xi in mpi_partition(comm,
                                       list(enumerate(x)),
                                       root=root):
                assert x[i] == xi
                x_accessed_local[i] += 1
            x_accessed = array.array('i',[0])*len(x)
            if comm is not None:
                comm.Allreduce([x_accessed_local,
                                mpi4py.MPI.INT],
                               [x_accessed,
                                mpi4py.MPI.INT],
                               op=mpi4py.MPI.SUM)
                comm.Barrier()
            else:
                x_accessed[:] = x_accessed_local[:]
            for xi in x_accessed:
                assert xi == 1

def test_mpi_partition_no_comm():
    _test_mpi_partition(None)

if mpi_available:

    @MPITest(commsize=[1, 2, 4])
    def test_mpi_partition(comm):
        _test_mpi_partition(comm)



#
# This example defines a script that solves the
# rosenbrock_2d example, utilizing parallel computation in
# two different ways. It first executes optimality-based
# range reduction in parallel for 7 nodes of a serial
# branch-and-bound solve. After this completes, it continues
# the solve using parallel branch-and-bound (without range
# reduction), initializing the dispatcher queue with the
# unprocessed nodes from the previous solve.
#
# This example can be executed in serial as
#
# $ python range_reduction_pyomo.py
#
# or in parallel as
#
# $ mpiexec -n <n> python range_reduction_pyomo.py
#
import os
import sys

examples_dir = os.path.dirname(
    os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, os.path.join(examples_dir,
                                'command_line_problems'))
from rosenbrock_2d import Rosenbrock2D
sys.path.remove(os.path.join(examples_dir,
                             'command_line_problems'))

import pybnb
from pybnb.pyomo.range_reduction import RangeReductionProblem

class Rosenbrock2D_RangeReduction(RangeReductionProblem):

    improved_abstol = 1e-3

    #
    # Implement RangeReductionProblem abstract methods
    #

    def range_reduction_model_setup(self):
        """Called prior to starting range reduction solves
        to set up the Pyomo model"""
        self.problem.setup_model_for_bound()

    def range_reduction_objective_changed(self, objective):
        """Called to notify that the range reduction routine
        has changed the objective"""
        # nothing to do for this problem
        pass

    def range_reduction_constraint_added(self, constraint):
        """Called to notify that the range reduction routine
        has added a constraint"""
        # nothing to do for this problem
        pass

    def range_reduction_constraint_removed(self, constraint):
        """Called to notify that the range reduction routine
        has removed a constraint"""
        # nothing to do for this problem
        pass

    def range_reduction_get_objects(self):
        """Called to collect the set of objects over which
        to perform range reduction solves"""
        vlist = []
        x, y = self.problem._model.x, self.problem._model.y
        if (y.ub - y.lb) > self.improved_abstol:
            vlist.append(y)
        if (x.ub - x.lb) > self.improved_abstol:
            vlist.append(x)
        return vlist

    def range_reduction_solve_for_object_bound(self, x):
        """Called to perform a range reduction solve for a
        Pyomo model object"""
        results = self.problem._ipopt.solve(self.problem._model,
                                            load_solutions=False)
        if str(results.solver.termination_condition) == "optimal":
            assert str(results.solver.status) == "ok"
            symbol_map = results.solution(0).symbol_map
            assert results.solution(0).default_variable_value is None
            return results.solution(0).\
                variable[symbol_map.getSymbol(x)]['Value']

    def range_reduction_model_cleanup(self):
        """Called after range reduction has finished to
        allow the user to execute any cleanup to the Pyomo
        model."""
        # nothing to do for this problem
        pass

    def range_reduction_process_bounds(self,
                                       objects,
                                       lower_bounds,
                                       upper_bounds):
        """Called to process the bounds obtained by the
        range reduction solves"""
        changed_cnt = 0
        for lb, ub, var in zip(lower_bounds,
                               upper_bounds,
                               objects):
            changed = False
            if lb > var.lb + self.improved_abstol:
                var.lb = lb
                changed = True
            if ub < var.ub - self.improved_abstol:
                var.ub = ub
                changed = True
            if changed:
                changed_cnt += 1
        self.problem.rebuild_convex_envelopes()
        return False

if __name__ == "__main__":
    import argparse

    comm = None
    try:
        import mpi4py.MPI
        comm = mpi4py.MPI.COMM_WORLD
    except ImportError:
        pass

    parser = argparse.ArgumentParser(
        description=("Run parallel branch and bound "
                     "with optimality-based range reduction "
                     "on the first few nodes"))
    parser.add_argument("--results-filename", type=str, default=None,
                        help=("When set, saves the solver results into a "
                              "YAML-formated file with the given name."))
    args = parser.parse_args()

    problem = Rosenbrock2D(xL=-25, xU=25,
                           yL=-25 ,yU=25)

    # do parallel bounds tightening on the
    # first 7 nodes that are processed
    obrr = Rosenbrock2D_RangeReduction(
        problem,
        comm=comm)

    queue = None
    best_node = None
    if (comm is None) or (comm.rank == 0):
        opt_obrr = pybnb.Solver(comm=None)
        results = opt_obrr.solve(obrr,
                                 node_limit=7,
                                 log_interval_seconds=0)
        queue = opt_obrr.save_dispatcher_queue()
        best_node = results.best_node
    else:
        obrr.listen(root=0)
    del obrr

    # continue the solve in parallel, without bounds
    # tightening on the remaining nodes
    results = pybnb.solve(problem,
                          comm=comm,
                          dispatcher_rank=0,
                          best_node=best_node,
                          initialize_queue=queue,
                          results_filename=args.results_filename)



#
# This example solves the bin packing problem. A custom
# heuristic is implemented to determine an integer-feasible
# objective value for each region explored. The objective
# heuristic is non-deterministic. Thus, this example may
# behave differently from run to run. Valid lower bounds are
# obtained by solving the linear relaxation using Pyomo.
#
# Pyomo version 5.4.3 or later is required to run this
# example.
#

import math
import random

import pybnb

pyomo_available = False
try:
    import pyomo.kernel as pmo
    if getattr(pmo,'version_info',(0,)*3) >= (5,4,3):  #pragma:nocover
        pyomo_available = True
except ImportError:                                    #pragma:nocover
    pass
if not pyomo_available:                                #pragma:nocover
    raise ImportError("This example requires Pyomo 5.4.3 or above")

class BinPacking(pybnb.Problem):

    def __init__(self, V, W,
                 pyomo_solver="ipopt",
                 pyomo_solver_io="nl",
                 integer_tolerance=1e-4):
        assert V > 0
        assert integer_tolerance > 0
        self.V = V
        self.W = W
        self._integer_tolerance = integer_tolerance
        N = range(len(self.W))
        m = self.model = pmo.block()
        x = m.x = pmo.variable_dict()
        y = m.y = pmo.variable_dict()
        for i in N:
            y[i] = pmo.variable(domain=pmo.Binary)
            for j in N:
                x[i,j] = pmo.variable(domain=pmo.Binary)

        m.B = pmo.expression(sum(y.values()))
        m.objective = pmo.objective(m.B, sense=pmo.minimize)

        m.B_nontrivial = pmo.constraint(m.B >= 1)

        m.capacity = pmo.constraint_dict()
        for i in N:
            m.capacity[i] = pmo.constraint(
                sum(x[i,j]*self.W[j] for j in N) <= self.V*y[i])

        m.assign_1 = pmo.constraint_dict()
        for j in N:
            m.assign_1[j] = pmo.constraint(
                sum(x[i,j] for i in N) == 1)

        # relax everything for the bound solves,
        # since the objective uses a simple heuristic
        self.true_domain_type = pmo.ComponentMap()
        for xij in self.model.x.components():
            self.true_domain_type[xij] = xij.domain_type
            xij.domain_type = pmo.RealSet
        for yi in self.model.y.components():
            self.true_domain_type[yi] = yi.domain_type
            yi.domain_type = pmo.RealSet

        self.opt = pmo.SolverFactory(
            pyomo_solver,
            solver_io=pyomo_solver_io)

    def _branch_y(self):

        N = range(len(self.W))
        for i in N:
            yi = self.model.y[i]
            assert yi.lb in (0,1), \
                str(yi.name)+" "+str(yi.bounds)
            assert yi.ub in (0,1), \
                str(yi.name)+" "+str(yi.bounds)
            if yi.lb == 0:
                if yi.ub == 1:
                    break
                else:
                    assert yi.ub == 0
            else:
                assert yi.lb == 1
                assert yi.lb == yi.ub
        else:
            # there is no branching left to do
            return ()

        for k in range(i,len(self.W)):
            assert self.model.y[k].lb == 0
            assert self.model.y[k].ub in (0,1)

        orig_bounds = pmo.ComponentMap()
        orig_bounds.update((yi, yi.bounds)
                           for yi in self.model.y.components())
        orig_bounds.update((xij, xij.bounds)
                           for xij in self.model.x.components())

        children = [pybnb.Node(),
                    pybnb.Node()]

        # first branch: fix this bin on
        self.model.y[i].lb = self.model.y[i].ub = 1
        self.save_state(children[0])

        # second branch: fix this bin off, as well as all
        #                bins following it
        for k in range(i,len(self.W)):
            self.model.y[k].lb = self.model.y[k].ub = 0
            for j in N:
                assert self.model.x[k,j].lb == 0
                self.model.x[k,j].ub = 0
        self.save_state(children[1])

        # reset bounds
        for var in orig_bounds:
            var.bounds = orig_bounds[var]

        return children

    def _branch_x(self):
        N = range(len(self.W))
        orig_bounds = pmo.ComponentMap()
        orig_bounds.update((yi, yi.bounds)
                           for yi in self.model.y.components())
        orig_bounds.update((xij, xij.bounds)
                           for xij in self.model.x.components())

        # do some simple preprocessing to reduce
        # the branch space
        for i in N:
            assert self.model.y[i].lb <= self.model.y[i].ub
            assert self.model.y[i].lb in (0,1)
            assert self.model.y[i].ub in (0,1)
            if self.model.y[i].ub == 0:
                for j in N:
                    assert self.model.x[i,j].lb == 0
                    self.model.x[i,j].ub = 0
        unassigned_items = []
        for j in N:
            assigned = sum(self.model.x[i,j].lb for i in N)
            if assigned == 0:
                unassigned_items.append(j)
            else:
                assert assigned == 1
                for i in N:
                    if self.model.x[i,j].lb == 0:
                        self.model.x[i,j].ub = 0
        # find an item that is not already fixed into a bin
        bv = None
        for j in unassigned_items:
            for i in N:
                if (self.model.x[i,j].lb == 0) and \
                   (self.model.x[i,j].ub == 1):
                    bv = self.model.x[i,j]
                    break
            if bv is not None:
                break
        else:                                     #pragma:nocover
            return ()

        assert bv is not None
        children = [pybnb.Node(),
                    pybnb.Node()]
        bv.lb = bv.ub = 1
        self.save_state(children[0])
        bv.lb = bv.ub = 0
        self.save_state(children[1])

        # reset bounds
        for var in orig_bounds:
            var.bounds = orig_bounds[var]

        return children

    #
    # Implement Problem abstract methods
    #

    def sense(self):
        return pybnb.minimize

    def objective(self):
        for xij in self.model.x.components():
            assert xij.lb in (0,1)
            assert xij.ub in (0,1)
            xij.value = xij.lb
        for yi in self.model.y.components():
            assert yi.lb in (0,1)
            assert yi.ub in (0,1)
            yi.value = yi.lb
        # a simple heuristic: iterate over items in random order
        #                     and fill up bins, adding as needed
        N = range(len(self.W))
        item_order = []
        for j in N:
            assigned = sum(bool(self.model.x[i,j].value == 1)
                           for i in N)
            if assigned == 1:
                for i in N:
                    if self.model.x[i,j].value == 1:
                        assert self.model.y[i].ub == 1
                        self.model.y[i].value = 1
            else:
                assert assigned == 0
                item_order.append(j)

        on_bins = []
        off_bins = []
        for i in N:
            yi = self.model.y[i]
            if yi.value == 1:
                on_bins.append(i)
            else:
                assert yi.value == 0
                off_bins.append(i)
        bins = on_bins + off_bins
        bin_weight = {i: sum(self.W[j]*self.model.x[i,j].value for j in N)
                      for i in N}

        random.shuffle(item_order)
        for j in item_order:
            assert self.W[j] > 0
            for i in bins:
                if self.model.x[i,j].ub == 1:
                    if self.model.y[i].value == 0:
                        assert self.model.y[i].ub == 1
                        self.model.y[i].value = 1
                    if bin_weight[i] + self.W[j] <= self.V:
                        assert self.model.x[i,j].value == 0
                        assert self.model.x[i,j].lb == 0
                        assert self.model.x[i,j].ub == 1
                        self.model.x[i,j].value = 1
                        bin_weight[i] += self.W[j]
                        break
            else:
                return self.infeasible_objective()

        return self.model.objective()

    def bound(self):
        results = self.opt.solve(self.model, load_solutions=False)
        if (str(results.solver.status) == "ok") and \
           (str(results.solver.termination_condition) == "optimal"):
            self.model.load_solution(results.solution(0))
            # note that any integer feasible solution will have an
            # integer objective value, so we can strengthen the bound
            objective = self.model.objective()
            if objective - math.floor(objective) > \
               self._integer_tolerance:
                return  math.ceil(objective)
            else:
                return math.floor(objective)
        elif str(results.solver.termination_condition) == "infeasible":
            assert str(results.solver.status) in ("ok","warning")
            return self.infeasible_objective()
        else:                                     #pragma:nocover
            assert str(results.solver.status) == "error"
            assert "Restoration Phase Failed" in \
                results.solver.message
            # assume infeasible
            return self.infeasible_objective()

    def save_state(self, node):
        node.state = []
        for v in (self.model.x,
                  self.model.y):
            for vi in v.components():
                node.state.append(vi.bounds)

    def load_state(self, node):
        k = 0
        for v in (self.model.x,
                  self.model.y):
            for vi in v.components():
                vi.bounds = node.state[k]
                k += 1

    def branch(self):
        # try to branch on y
        children = self._branch_y()
        if len(children):
            return children
        # otherwise, branch on x
        return self._branch_x()

if __name__ == "__main__":
    import pybnb.misc

    V = 100
    W = [49,41,34,33,29,26,26,22,20,19]

    #V = 524
    #W = [442,252,252,252,252,252,252,252,
    #     127,127,127,127,127,
    #     106,106,106,106,
    #     85,84,46,37,37,12,12,12,
    #     10,10,10,10,10,10,9,9]

    problem = BinPacking(V,W)
    pybnb.misc.create_command_line_solver(problem)



"""
Basic definitions and utilities.

Copyright by Gabriel A. Hackebeil (gabe.hackebeil@gmail.com).
"""

import enum

minimize = 1
"""The objective sense defining a minimization problem."""

maximize = -1
"""The objective sense defining a maximization problem."""

inf = float("inf")
"""A floating point constant set to ``float('inf')``."""

nan = float("nan")
"""A floating point constant set to ``float('nan')``."""

@enum.unique
class QueueStrategy(str, enum.Enum):
    """Strategies for prioritizing nodes in the central
    dispatcher queue. For all strategies, ties are broken by
    insertion order."""

    bound = "bound"
    """The node with the worst bound is always selected
    next."""
    objective = "objective"
    """The node with the best objective is always selected
    next."""
    breadth = "breadth"
    """The node with the smallest tree depth is always
    selected next (i.e., breadth-first search)."""
    depth = "depth"
    """The node with the largest tree depth is always
    selected next (i.e., depth-first search)."""
    local_gap = "local_gap"
    """The node with the largest gap between its local
    objective and bound is always selected next."""
    fifo = "fifo"
    """Nodes are served in first-in, first-out order."""
    lifo = "lifo"
    """Nodes are served in last-in, first-out order."""
    random = "random"
    """Nodes are assigned a random priority before entering
    the queue."""
    custom = "custom"
    """The node with the largest value stored in the
    :attr:`queue_priority <pybnb.node.Node.queue_priority>`
    attribute is always selected next. Users are expected to
    assign a priority to all nodes returned from the
    :func:`branch <pybnb.problem.Problem.branch>`
    method on their problem."""
_queue_strategy_to_int = {}
_int_to_queue_strategy = []
for _i, _val in enumerate(sorted(QueueStrategy,
                                 key=lambda v: v.value)):
    _queue_strategy_to_int[_val] = _i
    _int_to_queue_strategy.append(_val)
del _i
del _val

@enum.unique
class SolutionStatus(str, enum.Enum):
    """Possible values assigned to the
    :attr:`solution_status` attribute of a
    :class:`SolverResults <pybnb.solver_results.SolverResults>`
    object returned from a solve."""

    optimal = "optimal"
    """Indicates that the best objective is finite and close
    enough to the global bound to satisfy the optimality
    tolerances used for the solve."""
    feasible = "feasible"
    """Indicates that the best objective is finite but not
    close enough to the global bound to satisfy the
    optimality tolerances used for the solve."""
    infeasible = "infeasible"
    """Indicates that both the best objective and global
    bound are equal to the infeasible objective value (+inf
    or -inf depending on the sense)."""
    unbounded = "unbounded"
    """Indicates that both the best objective and global
    bound are equal to the unbounded objective value (+inf
    or -inf depending on the sense)."""
    invalid = "invalid"
    """Indicates that the global bound is not a valid bound
    on the best objective found. This may be due to an
    ill-defined problem or other numerical issues."""
    unknown = "unknown"
    """Indicates that the global bound is finite, but no
    feasible (finite) objective was found."""
_solution_status_to_int = {}
_int_to_solution_status = []
for _i, _val in enumerate(sorted(SolutionStatus,
                                 key=lambda v: v.value)):
    _solution_status_to_int[_val] = _i
    _int_to_solution_status.append(_val)
del _i
del _val

@enum.unique
class TerminationCondition(str, enum.Enum):
    """Possible values assigned to the
    :attr:`termination_condition` attribute of a
    :class:`SolverResults <pybnb.solver_results.SolverResults>`
    object returned from a solve."""

    optimality = "optimality"
    """The dispatcher terminated the solve based on
    optimality criteria."""
    objective_limit = "objective_limit"
    """The dispatcher terminated the solve based on the
    user-supplied limit on the objective or bound being
    satisfied."""
    node_limit = "node_limit"
    """The dispatcher terminated the solve due to the
    user-supplied explored node limit being surpassed."""
    time_limit = "time_limit"
    """The dispatcher terminated the solve due to the
    user-supplied time limit being surpassed."""
    no_nodes = "no_nodes"
    """The dispatcher terminated the solve due to the node
    queue becoming empty."""
    interrupted = "interrupted"
    """Solve termination was initiated by SIGINT or SIGUSR
    signal event."""
_termination_condition_to_int = {}
_int_to_termination_condition = []
for _i, _val in enumerate(sorted(TerminationCondition,
                                 key=lambda v: v.value)):
    _termination_condition_to_int[_val] = _i
    _int_to_termination_condition.append(_val)
del _i
del _val



"""
Branch-and-bound solver results object.

Copyright by Gabriel A. Hackebeil (gabe.hackebeil@gmail.com).
"""

# recognized pytest-doctestplus plugin,
# not the standard doctest
__doctest_requires__ = {'SolverResults.write': ['yaml']}

import sys
import base64

from pybnb.common import (SolutionStatus,
                          TerminationCondition)
from pybnb.misc import (time_format,
                        as_stream)
from pybnb.node import dumps

import six

class SolverResults(object):
    """Stores the results of a branch-and-bound solve.

    Attributes
    ----------
    solution_status : string
        The solution status will be set to one of the strings
        documented by the :class:`SolutionStatus
        <pybnb.common.SolutionStatus>` enum.
    termination_condition : string
        The solve termination condition, as determined by
        the dispatcher, will be set to one of the strings
        documented by the :class:`TerminationCondition
        <pybnb.common.TerminationCondition>` enum.
    objective : float
        The best objective found.
    bound : float
        The global optimality bound.
    absolute_gap : float
        The absolute gap between the objective and bound.
    relative_gap : float
        The relative gap between the objective and bound.
    nodes : float
        The total number of nodes processes by all workers.
    wall_time : float
        The process-local wall time (seconds). This is the
        only value on the results object that varies between
        processes.
    best_node : :class:`Node <pybnb.node.Node>`
        The node with the best objective obtained during the
        solve. Note that if the best_objective solver option
        was used, the best_node on the results object may
        have an objective that is worse than the objective
        stored on the results (or may be None).
    """

    def __init__(self):
        self.solution_status = None
        self.termination_condition = None
        self.objective = None
        self.bound = None
        self.absolute_gap = None
        self.relative_gap = None
        self.nodes = None
        self.wall_time = None
        self.best_node = None

    def pprint(self, stream=sys.stdout):
        """Prints a nicely formatted representation of the
        results.

        Parameters
        ----------
        stream : file-like object or string, optional
            A file-like object or a filename where results
            should be written to. (default: ``sys.stdout``)
        """
        with as_stream(stream) as stream:
            stream.write("solver results:\n")
            self.write(stream, prefix=" - ", pretty=True)

    def write(self, stream, prefix="", pretty=False):
        """Writes results in YAML format to a stream or
        file. Changing the parameter values from their
        defaults may result in the output becoming
        non-compatible with the YAML format.

        Parameters
        ----------
        stream : file-like object or string
            A file-like object or a filename where results
            should be written to.
        prefix : string, optional
            A string to use as a prefix for each line that
            is written. (default: '')
        pretty : bool, optional
            Indicates whether or not certain recognized
            attributes should be formatted for more
            human-readable output. (default: False)

        Example
        -------

        >>> import six
        >>> import yaml
        >>> import pybnb
        >>> results = pybnb.SolverResults()
        >>> results.best_node = pybnb.Node()
        >>> results.best_node.objective = 123
        >>> out = six.StringIO()
        >>> # the best_node is serialized
        >>> results.write(out)
        >>> del results
        >>> results_dict = yaml.load(out.getvalue())
        >>> # de-serialize the best_node
        >>> best_node = pybnb.node.loads(results_dict['best_node'])
        >>> assert best_node.objective == 123

        """
        with as_stream(stream) as stream:
            attrs = vars(self)
            names = sorted(list(attrs.keys()))
            first = ('solution_status', 'termination_condition',
                     'objective', 'bound',
                     'absolute_gap', 'relative_gap',
                     'nodes', 'wall_time', 'best_node')
            for cnt, name in enumerate(first):
                if not hasattr(self, name):
                    continue
                names.remove(name)
                val = getattr(self, name)
                if val is not None:
                    if name in ("solution_status",
                                "termination_condition"):
                        if type(val) in (SolutionStatus,
                                         TerminationCondition):
                            val = val.value
                    elif pretty:
                        if name == 'wall_time':
                            val = time_format(val,
                                              digits=2)
                        elif name in ('objective',
                                      'bound',
                                      'absolute_gap',
                                      'relative_gap'):
                            val = "%.7g" % (val)
                        elif name == "best_node":
                            if val.objective is not None:
                                val = ("Node(objective=%.7g)"
                                       % (val.objective))
                            else:
                                val = "Node(objective=None)"
                    else:
                        if name == "best_node":
                            val = dumps(val)
                            if hasattr(base64, 'encodebytes'):
                                val = base64.encodebytes(val).\
                                    decode("ascii")
                            else:
                                val = base64.encodestring(val).\
                                    decode("ascii")
                            val = ('\n  '.join(
                                val.splitlines()))
                            val = ("!!binary |\n  %s"
                                   % (val))
                        else:
                            val_ = "%r" % (val)
                            if type(val) is float:
                                if val_ == 'inf':
                                    val_ = '.inf'
                                elif val_ == '-inf':
                                    val_ = "-.inf"
                                elif val_ == 'nan':
                                    val_ = ".nan"
                            val = val_
                            del val_
                if pretty or (val is not None):
                    stream.write(prefix+'%s: %s\n'
                                 % (name, val))
                else:
                    assert val is None
                    stream.write(prefix+'%s: null\n'
                                 % (name))
            for name in names:
                val = getattr(self, name)
                if pretty:
                    stream.write(prefix+'%s: %r\n'
                                 % (name, val))
                else:
                    if val is None:
                        stream.write(prefix+'%s: null\n'
                                     % (name))
                    else:
                        val_ = "%r" % (val)
                        if type(val) is float:
                            if val_ == 'inf':
                                val_ = '.inf'
                            elif val_ == '-inf':
                                val_ = "-.inf"
                            elif val_ == 'nan':
                                val_ = ".nan"
                        val = val_
                        del val_
                        stream.write(prefix+'%s: %s\n'
                                     % (name, val))

    def __str__(self):
        """Represents the results as a string."""
        tmp = six.StringIO()
        self.pprint(stream=tmp)
        return tmp.getvalue()



#
# This example script defines and solves
# a minimal pybnb example problem.
#
# It can be executed in serial as
#
# $ python simple.py
#
# or in parallel as
#
# $ mpiexec -n <n> python simple.py
#
# The mpi4py module is required.
#
import pybnb

class Simple(pybnb.Problem):
    def __init__(self):
        self._xL, self._xU = 0, 1
    #
    # required methods
    #
    def sense(self):
        return pybnb.minimize
    def objective(self):
        return round(self._xU-self._xL,3)
    def bound(self):
        return -(self._xU - self._xL)**2
    def save_state(self, node):
        node.state = (self._xL, self._xU)
    def load_state(self, node):
        (self._xL, self._xU) = node.state
    def branch(self):
        xL, xU = self._xL, self._xU
        xM = 0.5 * (xL + xU)
        child = pybnb.Node()
        child.state = (xL, xM)
        yield child
        child = pybnb.Node()
        child.state = (xM, xU)
        yield child
    #
    # optional methods
    #
    def notify_solve_begins(self,
                            comm,
                            worker_comm,
                            convergence_checker):
        pass
    def notify_new_best_node(self,
                             node,
                             current):
        pass
    def notify_solve_finished(self,
                              comm,
                              worker_comm,
                              results):
        pass

problem = Simple()
solver = pybnb.Solver()
results = solver.solve(problem)



"""
A Problem interface for implementing parallel range
reduction on a PyomoProblem during a branch-and-bound solve.

Copyright by Gabriel A. Hackebeil (gabe.hackebeil@gmail.com).
"""
import math
import array

from pybnb import (inf, Problem)
from pybnb.node import Node
from pybnb.pyomo.misc import (hash_joblist,
                              add_tmp_component,
                              mpi_partition,
                              create_optimality_bound)
from pybnb.pyomo.problem import PyomoProblem

import pyomo.kernel as pmo

try:
    import mpi4py
except ImportError:                               #pragma:nocover
    pass

class RangeReductionProblem(Problem):
    """A specialized implementation of the
    :class:`pybnb.Problem <pybnb.problem.Problem>` interface
    that can be used to perform optimality-based range
    reduction on a fully implemented :class:`PyomoProblem`
    by defining additional abstract methods."""

    def __init__(self,
                 problem,
                 best_objective=None,
                 comm=None):
        assert isinstance(problem, PyomoProblem)
        self.problem = problem
        assert best_objective != self.unbounded_objective
        self._best_objective = best_objective
        if self._best_objective is None:
            self._best_objective = self.problem.infeasible_objective()
        assert not math.isnan(self._best_objective)
        self._comm = comm
        if self._comm is not None:
            import mpi4py.MPI
        self._current_node = None

    def _notify_continue_listen(self, node):
        assert (self._comm is not None) and \
            (self._comm.size > 1)
        data = array.array('d', [True,
                                 self._best_objective,
                                 len(node.state)])
        self._comm.Bcast([data, mpi4py.MPI.DOUBLE],
                         root=self._comm.rank)
        node.state = self._comm.bcast(node.state,
                                      root=self._comm.rank)

    def _notify_stop_listen(self):
        assert (self._comm is not None) and \
            (self._comm.size > 1)
        data = array.array('d',[False,
                                self._best_objective,
                                0])
        self._comm.Bcast([data, mpi4py.MPI.DOUBLE],
                         root=self._comm.rank)

    def _tighten_bounds(self):
        self.range_reduction_model_setup()
        assert self._best_objective != self.unbounded_objective()
        # setup objective
        assert self.problem.pyomo_model_objective.active
        self.problem.pyomo_model_objective.deactivate()
        tmp_objective = pmo.objective()
        tmp_objective_name = add_tmp_component(
            self.problem.pyomo_model,
            "rr_objective",
            tmp_objective)
        # setup optimality bound if necessary
        tmp_optbound_name = None
        tmp_optbound = None
        if self._best_objective != self.infeasible_objective():
            tmp_optbound = create_optimality_bound(
                self,
                self.problem.pyomo_model_objective,
                self._best_objective)
            tmp_optbound_name = add_tmp_component(
                self.problem.pyomo_model,
                "optimality_bound",
                tmp_optbound)
            self.range_reduction_constraint_added(tmp_optbound)
        try:
            return self._tighten_bounds_impl(tmp_objective)
        finally:
            # reset objective
            delattr(self.problem.pyomo_model, tmp_objective_name)
            self.problem.pyomo_model_objective.activate()
            self.range_reduction_objective_changed(
                self.problem.pyomo_model_objective)
            # remove optimality bound if it was added
            if tmp_optbound is not None:
                self.range_reduction_constraint_removed(tmp_optbound)
                delattr(self.problem.pyomo_model, tmp_optbound_name)
            self.range_reduction_model_cleanup()

    def _tighten_bounds_impl(self, tmp_objective):
        objlist = self.range_reduction_get_objects()
        joblist = []
        objects = []
        lower_bounds = []
        upper_bounds = []
        objects_seen = set()
        for i, val in enumerate(objlist):
            obj = None
            include = False
            val = val if type(val) is tuple else (val,True,True)
            assert len(val) == 3
            obj = val[0]
            cid = self.problem.pyomo_object_to_cid[obj]
            if val[1]:
                include = True
                joblist.append((i,cid,'L'))
            if val[2]:
                include = True
                joblist.append((i,cid,'U'))
                joblist.append((i,cid,'U'))
            if include:
                assert obj is not None
                assert id(obj) not in objects_seen
                objects_seen.add(id(obj))
                objects.append(obj)
                lower_bounds.append(pmo.value(obj.lb) \
                                    if obj.has_lb() else \
                                    -inf)
                upper_bounds.append(pmo.value(obj.ub) \
                                    if obj.has_ub() else \
                                    inf)
        lower_bounds = array.array('d', lower_bounds)
        upper_bounds = array.array('d', upper_bounds)

        # verify that everyone has the exact same list
        # (order and values), assumes everything in the list
        # has a well-defined hash
        if self._comm is not None:
            my_joblist_hash = hash_joblist(joblist)
            joblist_hash = self._comm.bcast(my_joblist_hash,
                                            root=0)
            assert joblist_hash == my_joblist_hash
        for i, cid, which in mpi_partition(self._comm,
                                           joblist):
            obj = self.problem.cid_to_pyomo_object[cid]
            tmp_objective.expr = obj
            if which == 'L':
                tmp_objective.sense = pmo.minimize
            else:
                assert which == 'U'
                tmp_objective.sense = pmo.maximize
            self.range_reduction_objective_changed(tmp_objective)
            bound = self.range_reduction_solve_for_object_bound(obj)
            if bound is not None:
                if which == 'L':
                    lower_bounds[i] = bound
                else:
                    assert which == 'U'
                    upper_bounds[i] = bound
        if self._comm is not None:
            lower_bounds_local = lower_bounds
            upper_bounds_local = upper_bounds
            lower_bounds = array.array('d', lower_bounds)
            upper_bounds = array.array('d', upper_bounds)
            self._comm.Allreduce([lower_bounds_local,
                                  mpi4py.MPI.DOUBLE],
                                 [lower_bounds,
                                  mpi4py.MPI.DOUBLE],
                                 op=mpi4py.MPI.MAX)
            self._comm.Allreduce([upper_bounds_local,
                                  mpi4py.MPI.DOUBLE],
                                 [upper_bounds,
                                  mpi4py.MPI.DOUBLE],
                                 op=mpi4py.MPI.MIN)

        return objects, lower_bounds, upper_bounds

    #
    # Interface
    #

    def listen(self, root=0):
        """Listen for requests to run range reduction. All
        processes within the communicator, except for the
        root process, should call this method.

        Parameters
        ----------
        root : int
            The rank of the process acting as the
            root. The root process should not call this
            function.
        """
        assert self._comm.size > 1
        assert self._comm.rank != root
        orig = Node()
        self.save_state(orig)
        node = Node()
        try:
            data = array.array('d',[0])*3
            self._comm.Bcast([data,mpi4py.MPI.DOUBLE],
                             root=root)
            again = bool(data[0])
            self._best_objective = float(data[1])
            while again:
                node.state = self._comm.bcast(node.state,
                                              root=root)
                self.load_state(node)
                self._tighten_bounds()
                self._comm.Bcast([data,mpi4py.MPI.DOUBLE],
                                 root=root)
                again = bool(data[0])
                self._best_objective = float(data[1])
        finally:
            self.load_state(orig)

    #
    # Implement Problem abstract methods
    #

    def sense(self):
        return self.problem.sense()

    def objective(self):
        return self.problem.objective()

    def bound(self):
        # tell the listeners to start bounds tightening
        node = Node()
        self.save_state(node)
        continue_loop = True
        while continue_loop:
            if (self._comm is not None) and \
               (self._comm.size > 1):
                self._notify_continue_listen(node)
            continue_loop = self.range_reduction_process_bounds(
                *self._tighten_bounds())
            self.save_state(node)
        return self.problem.bound()

    def save_state(self, node):
        self.problem.save_state(node)

    def load_state(self, node):
        self.problem.load_state(node)

    def branch(self):
        return self.problem.branch()

    def notify_new_best_node(self,
                             node,
                             current):
        self._best_objective = node.objective

    def notify_solve_finished(self,
                              comm,
                              worker_comm,
                              results):
        if (self._comm is not None) and \
           (self._comm.size > 1):
            self._notify_stop_listen()

    #
    # Abstract Methods
    #

    def range_reduction_model_setup(self):
        """Called prior to starting range reduction solves
        to set up the Pyomo model"""
        raise NotImplementedError()               #pragma:nocover

    def range_reduction_objective_changed(self, objective):
        """Called to notify that the range reduction routine
        has changed the objective"""
        raise NotImplementedError()               #pragma:nocover

    def range_reduction_constraint_added(self, constraint):
        """Called to notify that the range reduction routine
        has added a constraint"""
        raise NotImplementedError()               #pragma:nocover

    def range_reduction_constraint_removed(self, constraint):
        """Called to notify that the range reduction routine
        has removed a constraint"""
        raise NotImplementedError()               #pragma:nocover

    def range_reduction_get_objects(self):
        """Called to collect the set of objects over which
        to perform range reduction solves"""
        raise NotImplementedError()               #pragma:nocover

    def range_reduction_solve_for_object_bound(self, x):
        """Called to perform a range reduction solve for a
        Pyomo model object"""
        raise NotImplementedError()               #pragma:nocover

    def range_reduction_model_cleanup(self):
        """Called after range reduction has finished to
        allow the user to execute any cleanup to the Pyomo
        model."""
        raise NotImplementedError()               #pragma:nocover

    def range_reduction_process_bounds(self,
                                       objects,
                                       lower_bounds,
                                       upper_bounds):
        """Called to process the bounds obtained by the
        range reduction solves"""
        raise NotImplementedError()               #pragma:nocover
